from importlib import import_module
from typing import Callable

import numpy as np


def from_yearly(value, time_step):
    return value * (time_step.total_seconds() / (60*60*24*365.0))


def to_yearly(value, time_step):
    return value / (time_step.total_seconds() / (60*60*24*365.0))


def rate_to_probability(rate):
    return 1-np.exp(-rate)


def probability_to_rate(probability):
    return -np.log(1-probability)


def collapse_nested_dict(d, prefix=None):
    results = []
    for k, v in d.items():
        cur_prefix = prefix+'.'+k if prefix else k
        if isinstance(v, dict):
            results.extend(collapse_nested_dict(v, prefix=cur_prefix))
        else:
            results.append((cur_prefix, v))
    return results


def import_by_path(path: str) -> Callable:
    """Import a class or function given it's absolute path.

    Parameters
    ----------
    path:
      Path to object to import
    """

    module_path, _, class_name = path.rpartition('.')
    return getattr(import_module(module_path), class_name)



import pytest

from vivarium.framework.engine import SimulationContext, Builder, setup_simulation, run, run_simulation
from vivarium.framework.event import EventManager, EventInterface
from vivarium.framework.lookup import InterpolatedDataManager, LookupTableInterface
from vivarium.framework.components import ComponentManager, ComponentInterface
from vivarium.framework.metrics import Metrics
from vivarium.framework.plugins import PluginManager
from vivarium.framework.population import PopulationManager, PopulationInterface
from vivarium.framework.randomness import RandomnessManager, RandomnessInterface
from vivarium.framework.time import DateTimeClock, TimeInterface
from vivarium.framework.values import ValuesManager, ValuesInterface

from .components.mocks import MockComponentA, MockComponentB, Listener


def is_same_object_method(m1, m2):
    return m1.__func__ is m2.__func__ and m1.__self__ is m2.__self__


@pytest.fixture
def components():
    return [MockComponentA('gretchen', 'whimsy'),
            Listener('listener'),
            MockComponentB('spoon', 'antelope', 23)]


@pytest.fixture
def log(mocker):
    return mocker.patch('vivarium.framework.engine._log')


def test_SimulationContext_init_default(base_config, components):
    sim = SimulationContext(base_config, components)

    assert sim.configuration == base_config
    assert isinstance(sim.component_manager, ComponentManager)
    assert isinstance(sim.clock, DateTimeClock)
    assert isinstance(sim.values, ValuesManager)
    assert isinstance(sim.events, EventManager)
    assert isinstance(sim.population, PopulationManager)
    assert isinstance(sim.tables, InterpolatedDataManager)
    assert isinstance(sim.randomness, RandomnessManager)

    assert isinstance(sim.builder, Builder)
    assert sim.builder.configuration is sim.configuration
    assert isinstance(sim.builder.lookup, LookupTableInterface)
    assert sim.builder.lookup._lookup_table_manager is sim.tables
    assert isinstance(sim.builder.value, ValuesInterface)
    assert sim.builder.value._value_manager is sim.values
    assert isinstance(sim.builder.event, EventInterface)
    assert sim.builder.event._event_manager is sim.events
    assert isinstance(sim.builder.population, PopulationInterface)
    assert sim.builder.population._population_manager is sim.population
    assert isinstance(sim.builder.randomness, RandomnessInterface)
    assert sim.builder.randomness._randomness_manager is sim.randomness
    assert isinstance(sim.builder.time, TimeInterface)
    assert sim.builder.time._clock is sim.clock
    assert isinstance(sim.builder.components, ComponentInterface)
    assert sim.builder.components._component_manager is sim.component_manager

    # Ordering matters.
    managers = [sim.clock, sim.population, sim.randomness, sim.values, sim.events, sim.tables]
    assert sim.component_manager._managers == managers
    assert sim.component_manager._components[:-1] == components
    assert isinstance(sim.component_manager._components[-1], Metrics)


def test_SimulationContext_init_custom(base_config, components):
    beehive = MockComponentA('COVERED IN BEES')
    beekeeper = MockComponentB("Gets the honey")

    def plugin_interfaces_mock():
        return {'beehive': beehive}

    def plugin_controllers_mock():
        return {'beekeeper': beekeeper}

    plugin_manager = PluginManager()
    plugin_manager.get_optional_interfaces = plugin_interfaces_mock
    plugin_manager.get_optional_controllers = plugin_controllers_mock
    sim = SimulationContext(base_config, components, plugin_manager)

    assert sim.configuration == base_config
    assert isinstance(sim.component_manager, ComponentManager)
    assert isinstance(sim.clock, DateTimeClock)
    assert isinstance(sim.values, ValuesManager)
    assert isinstance(sim.events, EventManager)
    assert isinstance(sim.population, PopulationManager)
    assert isinstance(sim.tables, InterpolatedDataManager)
    assert isinstance(sim.randomness, RandomnessManager)

    assert isinstance(sim.builder, Builder)
    assert sim.builder.configuration is sim.configuration
    assert isinstance(sim.builder.lookup, LookupTableInterface)
    assert sim.builder.lookup._lookup_table_manager is sim.tables
    assert isinstance(sim.builder.value, ValuesInterface)
    assert sim.builder.value._value_manager is sim.values
    assert isinstance(sim.builder.event, EventInterface)
    assert sim.builder.event._event_manager is sim.events
    assert isinstance(sim.builder.population, PopulationInterface)
    assert sim.builder.population._population_manager is sim.population
    assert isinstance(sim.builder.randomness, RandomnessInterface)
    assert sim.builder.randomness._randomness_manager is sim.randomness
    assert isinstance(sim.builder.time, TimeInterface)
    assert sim.builder.time._clock is sim.clock
    assert isinstance(sim.builder.components, ComponentInterface)
    assert sim.builder.components._component_manager is sim.component_manager
    assert sim.builder.beehive == beehive

    # Ordering matters.
    managers = [sim.clock, sim.population, sim.randomness, sim.values, sim.events, sim.tables, beekeeper]
    assert sim.component_manager._managers == managers
    assert sim.component_manager._components[:-1] == components
    assert isinstance(sim.component_manager._components[-1], Metrics)


def test_SimulationContext_setup_default(base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if 'listener' in c.args][0]
    assert not listener.post_setup_called
    sim.setup()

    def unpack(component):
        if isinstance(component, MockComponentB) and len(component.args) > 1:
            return [component] + [MockComponentB(arg) for arg in component.args]
        return [component]
    unpacked_components = [c for component in components for c in unpack(component)]
    unpacked_components.insert(len(components), Metrics())
    for a, b in zip(sim.component_manager._components, unpacked_components):
        assert type(a) == type(b)
        if hasattr(a, 'args'):
            assert a.args == b.args

    assert is_same_object_method(sim.simulant_creator, sim.population._create_simulants)
    assert sim.time_step_events == ['time_step__prepare', 'time_step', 'time_step__cleanup', 'collect_metrics']
    for k in sim.time_step_emitters.keys():
        assert is_same_object_method(sim.time_step_emitters[k], sim.events._event_types[k].emit)

    assert is_same_object_method(sim.end_emitter, sim.events._event_types['simulation_end'].emit)

    assert listener.post_setup_called


def test_SimulationContext_step(log, base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()

    current_time = sim.clock.time
    step_size = sim.clock.step_size

    listener = [c for c in components if 'listener' in c.args][0]

    assert not listener.time_step__prepare_called
    assert not listener.time_step_called
    assert not listener.time_step__cleanup_called
    assert not listener.collect_metrics_called

    sim.step()

    assert log.debug.called_once_with(current_time)
    assert listener.time_step__prepare_called
    assert listener.time_step_called
    assert listener.time_step__cleanup_called
    assert listener.collect_metrics_called

    assert sim.clock.time == current_time + step_size


def test_SimulationContext_initialize_simulants(base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    pop_size = sim.configuration.population.population_size
    current_time = sim.clock.time
    assert sim.population.population.empty
    sim.initialize_simulants()
    assert len(sim.population.population) == pop_size
    assert sim.clock.time == current_time


def test_SimulationContext_finalize(base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if 'listener' in c.args][0]
    sim.setup()
    assert not listener.simulation_end_called
    sim.finalize()
    assert listener.simulation_end_called


def test_SimulationContext_report(base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    sim.initialize_simulants()
    metrics = sim.report()
    assert metrics['test'] == len([c for c in sim.component_manager._components if isinstance(c, MockComponentB)])


def test_setup_simulation(model_specification, mocker):
    plugin_manager_constructor_mock = mocker.patch('vivarium.framework.engine.PluginManager')
    plugin_manager_mock = mocker.MagicMock()
    plugin_manager_constructor_mock.return_value = plugin_manager_mock
    context_constructor_mock = mocker.patch('vivarium.framework.engine.SimulationContext')
    context_mock = mocker.MagicMock()
    context_constructor_mock.return_value = context_mock
    component_config_parser = mocker.MagicMock()
    component_config_parser.get_components.return_value = ['test']
    plugin_manager_mock.get_plugin.return_value = component_config_parser

    setup_simulation(model_specification)

    plugin_manager_constructor_mock.assert_called_once_with(model_specification.plugins)
    plugin_manager_mock.get_plugin.assert_called_once_with('component_configuration_parser')
    component_config_parser.get_components.assert_called_once_with(model_specification.components)
    context_constructor_mock.assert_called_once_with(model_specification.configuration, ['test'], plugin_manager_mock)
    context_mock.setup.assert_called_once()


def test_run(mocker):
    sim_mock = mocker.Mock()
    sim_mock.clock.time = 0
    sim_mock.clock.stop_time = 10

    def step():
        sim_mock.clock.time += 1

    sim_mock.step.side_effect = step
    sim_mock.report.return_value = {}
    pop = list(range(10))
    sim_mock.population.population = pop

    metrics, population = run(sim_mock)

    sim_mock.initialize_simulants.assert_called_once()
    assert sim_mock.step.call_count == sim_mock.clock.stop_time
    assert sim_mock.clock.time == sim_mock.clock.stop_time
    sim_mock.finalize.assert_called_once()
    sim_mock.report.assert_called_once()
    assert list(metrics.keys()) == ['simulation_run_time']
    assert metrics['simulation_run_time'] > 0
    assert pop is population


def test_run_simulation(model_specification, mocker):
    get_results_writer_mock = mocker.patch('vivarium.framework.engine.get_results_writer')
    results_writer_mock = mocker.Mock()
    results_writer_mock.results_root = 'test_dir'
    get_results_writer_mock.return_value = results_writer_mock

    build_model_spec_mock = mocker.patch('vivarium.framework.engine.build_model_specification')
    build_model_spec_mock.return_value = model_specification

    setup_simulation_mock = mocker.patch('vivarium.framework.engine.setup_simulation')
    sim_mock = mocker.Mock()
    unused_keys = ['not', 'used', 'at', 'all']
    sim_mock.configuration.unused_keys.return_value = unused_keys
    setup_simulation_mock.return_value = sim_mock

    run_mock = mocker.patch('vivarium.framework.engine.run')
    metrics = {'test_metrics': 'test'}
    final_state = 'the_final_state'
    run_mock.return_value = metrics, final_state

    log_mock = mocker.patch('vivarium.framework.engine._log')
    pformat_mock = mocker.patch('vivarium.framework.engine.pformat')
    pformat_mock.side_effect = lambda x: x

    pandas_mock = mocker.patch('vivarium.framework.engine.pd')
    pandas_mock.DataFrame.side_effect = lambda x, index: x

    model_spec_path = '/this/is/a/test.yaml'
    results_directory = 'test_dir'

    assert 'output_data' not in model_specification.configuration

    run_simulation(model_spec_path, results_directory)

    get_results_writer_mock.assert_called_once_with(results_directory, model_spec_path)

    build_model_spec_mock.assert_called_once_with(model_spec_path)
    assert model_specification.configuration.output_data.results_directory == results_directory

    setup_simulation_mock.assert_called_once_with(model_specification)
    run_mock.assert_called_once_with(sim_mock)

    pformat_mock.assert_called_once_with(metrics)
    log_mock.debug.assert_has_calls([mocker.call(metrics),
                                     mocker.call("Some configuration keys not used during run: %s", unused_keys)])

    results_writer_mock.write_output.assert_has_calls([mocker.call(metrics, 'output.hdf'),
                                                       mocker.call(final_state, 'final_state.hdf')])



"""The mutable value system"""
from collections import defaultdict
from typing import Callable
from types import MethodType
import logging

import pandas as pd

from vivarium import VivariumError
from .utilities import from_yearly


_log = logging.getLogger(__name__)


class DynamicValueError(VivariumError):
    """Indicates that a value was invoked without being properly configured.
    i.e. no source specified.
    """
    pass


def replace_combiner(value, mutator, *args, **kwargs):
    """Replaces the output of the source or mutator with the output
    of the subsequent mutator. This is the default combiner.
    """
    args = list(args) + [value]
    return mutator(*args, **kwargs)


def set_combiner(value, mutator, *args, **kwargs):
    """Expects the output of the source to be a set to which
    the result of each mutator is added.
    """
    value.add(mutator(*args, **kwargs))
    return value


def list_combiner(value, mutator, *args, **kwargs):
    """Expects the output of the source to be a list to which
    the result of each mutator is appended.
    """
    value.append(mutator(*args, **kwargs))
    return value


def rescale_post_processor(a, time_step):
    """Assumes that the value is an annual rate and rescales it to the
    current time step.
    """
    return from_yearly(a, pd.Timedelta(time_step, unit='D'))


def joint_value_post_processor(a, _):
    """The final step in calculating joint values like disability weights.
    If the combiner is list_combiner then the effective formula is:

    .. math::

        value(args) = 1 -  \prod_{i=1}^{mutator count} 1-mutator_{i}(args)

    Parameters
    ----------
    a : List[pd.Series]
        a is a list of series, indexed on the population. Each series
        corresponds to a different value in the pipeline and each row
        in a series contains a value that applies to a specific simulant.
    """
    # if there is only one value, return the value
    if len(a) == 1:
        return a[0]

    # if there are multiple values, calculate the joint value
    product = 1
    for v in a:
        new_value = (1-v)
        product = product * new_value
    joint_value = 1 - product
    return joint_value


class Pipeline:
    """A single mutable value."""

    def __init__(self):
        self.name = None
        self.source = None
        self.mutators = [[] for _ in range(10)]
        self.combiner = None
        self.post_processor = None
        self.manager = None
        self.configured = False

    def __call__(self, *args, skip_post_processor=False, **kwargs):
        if not self.source:
            raise DynamicValueError(f"The dynamic value pipeline for {self.name} has no source. This likely means"
                                    f"you are attempting to modify a value that hasn't been created.")
        elif not self.configured:
            raise DynamicValueError(f"The dynamic value pipeline for {self.name} has a source but "
                                    f"has not been configured.  You've done a weird thing to get in this state.")

        value = self.source(*args, **kwargs)
        for priority_bucket in self.mutators:
            for mutator in priority_bucket:
                value = self.combiner(value, mutator, *args, **kwargs)
        if self.post_processor and not skip_post_processor:
            return self.post_processor(value, self.manager.step_size())

        return value

    def __repr__(self):
        return f"_Pipeline({self.name})"


class ValuesManager:
    """The configuration of the dynamic values system.

    Notes
    -----
    Client code should never need to interact with this class
    except through the dynamic value and rate constructors exposed
    via the builder during the setup phase.
    """

    def __init__(self):
        self._pipelines = defaultdict(Pipeline)

    def setup(self, builder):
        self.step_size = builder.time.step_size()
        builder.event.register_listener('post_setup', self.on_post_setup)

    def on_post_setup(self, event):
        _log.debug(f"{[p for p, v in self._pipelines.items() if not v.source]}")

    def register_value_modifier(self, value_name, modifier, priority=5):
        m = modifier if isinstance(modifier, MethodType) else modifier.__call__
        _log.debug(f"Registering {str(m).split()[2]} as modifier to {value_name}")
        pipeline = self._pipelines[value_name]
        pipeline.mutators[priority].append(modifier)

    def register_value_producer(self, value_name, source=None,
                                preferred_combiner=replace_combiner, preferred_post_processor=None):
        _log.debug(f"Registering value pipeline {value_name}")
        pipeline = self._pipelines[value_name]
        pipeline.name = value_name
        pipeline.source = source
        pipeline.combiner = preferred_combiner
        pipeline.post_processor = preferred_post_processor
        pipeline.manager = self
        pipeline.configured = True
        return pipeline

    def register_rate_producer(self, rate_name, source=None):
        return self.register_value_producer(rate_name, source, preferred_post_processor=rescale_post_processor)

    def get_value(self, name):
        if name not in self._pipelines:
            raise DynamicValueError(f"The dynamic value {name} has not been registered with the value system.")
        return self._pipelines[name]

    def get_rate(self, name):
        if name not in self._pipelines:
            raise DynamicValueError(f"The dynamic rate {name} has not been registered with the value system.")
        return self._pipelines[name]

    def __contains__(self, item):
        return item in self._pipelines

    def __iter__(self):
        return iter(self._pipelines)

    def keys(self):
        return self._pipelines.keys()

    def items(self):
        return self._pipelines.items()

    def __repr__(self):
        return "ValuesManager(_pipelines= {})".format(list(self._pipelines.keys()))


class ValuesInterface:
    def __init__(self, value_manager: ValuesManager):
        self._value_manager = value_manager

    def register_value_producer(self, value_name: str, source: Callable[..., pd.DataFrame]=None,
                                preferred_combiner: Callable=replace_combiner,
                                preferred_post_processor: Callable[..., pd.DataFrame]=None) -> Pipeline:
        """Marks a ``Callable`` as the producer of a named value.

        Parameters
        ----------
        value_name :
            The name of the new dynamic value pipeline.
        source :
            A callable source for the dynamic value pipeline.
        preferred_combiner :
            A strategy for combining the source and the results of any calls to mutators in the pipeline.
            ``vivarium`` provides the strategies ``replace_combiner`` (the default), ``list_combiner``,
            and ``set_combiner`` which are importable from ``vivarium.framework.values``.  Client code
            may define additional strategies as necessary.
        preferred_post_processor :
            A strategy for processing the final output of the pipeline. ``vivarium`` provides the strategies
            ``rescale_post_processor`` and ``joint_value_post_processor`` which are importable from
            ``vivarium.framework.values``.  Client code may define additional strategies as necessary.

        Returns
        -------
        Callable
            A callable reference to the named dynamic value pipeline.
        """
        return self._value_manager.register_value_producer(value_name, source,
                                                           preferred_combiner,
                                                           preferred_post_processor)

    def register_rate_producer(self, rate_name: str, source: Callable[..., pd.DataFrame]=None) -> Pipeline:
        """Marks a ``Callable`` as the producer of a named rate.

        This is a convenience wrapper around ``register_value_producer`` that makes sure
        rate data is appropriately scaled to the size of the simulation time step.
        It is equivalent to ``register_value_producer(value_name, source,
        preferred_combiner=replace_combiner, preferred_post_processor=rescale_post_processor)``

        Parameters
        ----------
        rate_name :
            The name of the new dynamic rate pipeline.
        source :
            A callable source for the dynamic rate pipeline.

        Returns
        -------
        Callable
            A callable reference to the named dynamic rate pipeline.
        """
        return self._value_manager.register_rate_producer(rate_name, source)

    def register_value_modifier(self, value_name: str, modifier: Callable, priority: int=5):
        """Marks a ``Callable`` as the modifier of a named value.

        Parameters
        ----------
        value_name :
            The name of the dynamic value pipeline to be modified.
        modifier :
            A function that modifies the source of the dynamic value pipeline when called.
            If the pipeline has a ``replace_combiner``, the modifier should accept the same
            arguments as the pipeline source with an additional last positional argument
            for the results of the previous stage in the pipeline. For the ``list_combiner`` and
            ``set_combiner`` strategies, the pipeline modifiers should have the same signature
            as the pipeline source.
        priority : {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
            An indication of the order in which pipeline modifiers should be called. Modifiers with
            smaller priority values will be called earlier in the pipeline.  Modifiers with
            the same priority have no guaranteed ordering, and so should be commutative.
        """
        self._value_manager.register_value_modifier(value_name, modifier, priority)

    def get_value(self, name):
        return self._value_manager.get_value(name)



import pytest

from vivarium.config_tree import ConfigTree


def test_single_layer():
    d = ConfigTree()
    d.test_key = 'test_value'
    d.test_key2 = 'test_value2'

    assert d.test_key == 'test_value'
    assert d.test_key2 == 'test_value2'

    d.test_key2 = 'test_value3'

    assert d.test_key2 == 'test_value3'
    assert d.test_key == 'test_value'


def test_dictionary_style_access():
    d = ConfigTree()
    d.test_key = 'test_value'
    d['test_key2'] = 'test_value2'

    assert d['test_key'] == 'test_value'
    assert d.test_key2 == 'test_value2'


def test_get_missing():
    d = ConfigTree()
    d.test_key = 'test_value'

    # Missing keys should be empty containers
    assert len(d.missing_key) == 0


def test_multiple_layer_get():
    d = ConfigTree(layers=['first', 'second', 'third'])
    d._set_with_metadata('test_key', 'test_value', 'first')
    d._set_with_metadata('test_key', 'test_value2', 'second')
    d._set_with_metadata('test_key', 'test_value3', 'third')

    d._set_with_metadata('test_key2', 'test_value4', 'first')
    d._set_with_metadata('test_key2', 'test_value5', 'second')

    d._set_with_metadata('test_key3', 'test_value6', 'first')

    assert d.test_key == 'test_value3'
    assert d.test_key2 == 'test_value5'
    assert d.test_key3 == 'test_value6'


def test_outer_layer_set():
    d = ConfigTree(layers=['inner', 'outer'])
    d._set_with_metadata('test_key', 'test_value', 'inner')
    d._set_with_metadata('test_key', 'test_value2', 'outer')

    d.test_key = 'test_value3'

    assert d.test_key == 'test_value3'


def test_read_dict():
    d = ConfigTree(layers=['inner', 'outer'])
    d._read_dict({'test_key': 'test_value', 'test_key2': 'test_value2'}, layer='inner')
    d._read_dict({'test_key': 'test_value3'}, layer='outer')

    assert d.test_key == 'test_value3'
    assert d.test_key2 == 'test_value2'


def test_read_dict_nested():
    d = ConfigTree(layers=['inner', 'outer'])
    d._read_dict({'test_container': {'test_key': 'test_value', 'test_key2': 'test_value2'}}, layer='inner')
    d._read_dict({'test_container': {'test_key': 'test_value3'}}, layer='inner')

    assert d.test_container.test_key == 'test_value3'
    assert d.test_container.test_key2 == 'test_value2'

    d._read_dict({'test_container': {'test_key2': 'test_value4'}}, layer='outer')

    assert d.test_container.test_key2 == 'test_value4'


def test_source_metadata():
    d = ConfigTree(layers=['inner', 'outer'])
    d._read_dict({'test_key': 'test_value'}, layer='inner', source='initial_load')
    d._read_dict({'test_key': 'test_value2'}, layer='outer', source='update')

    assert d.metadata('test_key') == [
        {'layer': 'inner', 'default': False, 'source': 'initial_load', 'value': 'test_value'},
        {'layer': 'outer', 'default': True, 'source': 'update', 'value': 'test_value2'}]


def test_exception_on_source_for_missing_key():
    d = ConfigTree(layers=['inner', 'outer'])
    d._read_dict({'test_key': 'test_value'}, layer='inner', source='initial_load')

    with pytest.raises(KeyError) as excinfo:
        d.metadata('missing_key')
    assert 'missing_key' in str(excinfo.value)


def test_drop_layer():
    d = ConfigTree(layers=['a', 'b', 'c'])
    d._set_with_metadata('test_key', 'test_value', 'a')
    d._set_with_metadata('test_key', 'test_value2', 'b')
    d._set_with_metadata('test_key', 'test_value3', 'c')

    assert d.test_key == 'test_value3'
    d.drop_layer('c')
    assert d.test_key == 'test_value2'

    with pytest.raises(ValueError):
        d.drop_layer('c')


def test_reset_layer():
    d = ConfigTree(layers=['a', 'b', 'c'])
    d._set_with_metadata('test_key', 'test_value', 'a')
    d._set_with_metadata('test_key', 'test_value2', 'b')

    assert d.test_key == 'test_value2'
    d.reset_layer('b')
    assert d.test_key == 'test_value'
    d._set_with_metadata('test_key', 'test_value3', 'b')


def test_reset_layer_with_preserved_keys():
    d = ConfigTree(layers=['a', 'b', 'c'])
    d._set_with_metadata('test_key', 'test_value', 'a')
    d._set_with_metadata('test_key2', 'test_value2', 'a')
    d._set_with_metadata('test_key3', 'test_value3', 'a')
    d._set_with_metadata('test_key4', 'test_value4', 'a')
    d._set_with_metadata('test_key', 'test_value5', 'b')
    d._set_with_metadata('test_key2', 'test_value6', 'b')
    d._set_with_metadata('test_key3', 'test_value7', 'b')
    d._set_with_metadata('test_key4', 'test_value8', 'b')

    d.reset_layer('b', preserve_keys=['test_key2', 'test_key3'])
    assert d.test_key == 'test_value'
    assert d.test_key2 == 'test_value6'
    assert d.test_key3 == 'test_value7'
    assert d.test_key4 == 'test_value4'


def test_reset_layer_with_preserved_keys_at_depth():
    d = ConfigTree(layers=['a', 'b', 'c'])
    d._read_dict({
        'test_key': {
            'test_key2': 'test_value',
            'test_key3': {'test_key4': 'test_value2'}
        },
        'test_key5': {
            'test_key6': 'test_value3',
            'test_key7': 'test_value4'}
    }, layer='a')
    d._read_dict({
        'test_key': {
            'test_key2': 'test_value5',
            'test_key3': {'test_key4': 'test_value6'}
        },
        'test_key5': {
            'test_key6': 'test_value7',
            'test_key7': 'test_value8',
        }
    }, layer='b')

    d.reset_layer('b', preserve_keys=['test_key.test_key3', 'test_key5.test_key6'])
    assert d.test_key.test_key2 == 'test_value'
    assert d.test_key.test_key3.test_key4 == 'test_value6'
    assert d.test_key5.test_key6 == 'test_value7'
    assert d.test_key5.test_key7 == 'test_value4'


def test_unused_keys():
    d = ConfigTree({'test_key': {'test_key2': 'test_value', 'test_key3': 'test_value2'}})

    assert d.unused_keys() == {'test_key.test_key2', 'test_key.test_key3'}

    _ = d.test_key.test_key2

    assert d.unused_keys() == {'test_key.test_key3'}

    _ = d.test_key.test_key3

    assert not d.unused_keys()



import functools

from vivarium import VivariumError
from vivarium.framework.configuration import validate_model_specification_file


def verify_yaml(_, __, file_path: str) -> str:
    """Ensures the provided file exists and is a yaml file"""
    return validate_model_specification_file(file_path)


def run_from_ipython() -> bool:
    """Taken from https://stackoverflow.com/questions/5376837/how-can-i-do-an-if-run-from-ipython-test-in-python"""
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def log_progress(sequence, every=None, size=None, name='Items'):
    """Taken from https://github.com/alexanderkuk/log-progress"""
    from ipywidgets import IntProgress, HTML, VBox
    from IPython.display import display

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)     # every 0.5%
    else:
        assert every is not None, 'sequence is iterator, set every'

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = 'info'
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    display(box)

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = '{name}: {index} / ?'.format(
                        name=name,
                        index=index
                    )
                else:
                    progress.value = index
                    label.value = u'{name}: {index} / {size}'.format(
                        name=name,
                        index=index,
                        size=size
                    )
            yield record
    except Exception as e:
        progress.bar_style = 'danger'
        raise
    else:
        progress.bar_style = 'success'
        progress.value = index
        label.value = "{name}: {index}".format(
            name=name,
            index=str(index or '?')
        )


class InteractiveError(VivariumError):
    """Error raised when the Interactive context is in an inconsistent state."""
    pass


def raise_if_not_setup(system_type):
    type_error_map = {
        'run': 'Simulation must be setup before it can be run',
        'value': 'Value pipeline configuration is not complete until the simulation is setup.',
        'event': 'Event configuration is not complete until the simulation is setup.',
        'component': 'Component configuration is not complete until the simulation is setup.'
    }
    err_msg = type_error_map[system_type]

    def method_wrapper(context_method):

        @functools.wraps(context_method)
        def wrapped_method(*args, **kwargs):
            instance = args[0]
            if not instance._setup:
                raise InteractiveError(err_msg)
            return context_method(*args, **kwargs)

        return wrapped_method

    return method_wrapper



import os

import numpy as np
import pandas as pd

from vivarium.framework import randomness


class NonCRNTestPopulation:

    configuration_defaults = {
        'population': {
            'age_start': 0,
            'age_end': 100,
            'population_size': 100,
        },
        'input_data': {
            'location': "Kenya",
        }
    }

    def setup(self, builder):
        self.config = builder.configuration
        self.randomness = builder.randomness.get_stream('population_age_fuzz')
        columns = ['age', 'sex', 'location', 'alive', 'entrance_time', 'exit_time']
        self.population_view = builder.population.get_view(columns)

        builder.population.initializes_simulants(self.generate_test_population,
                                                 creates_columns=columns)

        builder.event.register_listener('time_step', self.age_simulants)

    def generate_test_population(self, pop_data):
        age_start = pop_data.user_data.get('age_start', self.config.population.age_start)
        age_end = pop_data.user_data.get('age_end', self.config.population.age_end)
        location = self.config.input_data.location

        population = _non_crn_build_population(pop_data.index, age_start, age_end, location,
                                               pop_data.creation_time, pop_data.creation_window, self.randomness)
        self.population_view.update(population)

    def age_simulants(self, event):
        population = self.population_view.get(event.index, query="alive == 'alive'")
        population['age'] += event.step_size / pd.Timedelta(days=365)
        self.population_view.update(population)


class TestPopulation(NonCRNTestPopulation):
    def setup(self, builder):
        super().setup(builder)
        self.age_randomness = builder.randomness.get_stream('age_initialization', for_initialization=True)
        self.register = builder.randomness.register_simulants

    def generate_test_population(self, pop_data):
        age_start = pop_data.user_data.get('age_start', self.config.population.age_start)
        age_end = pop_data.user_data.get('age_end', self.config.population.age_end)
        age_draw = self.age_randomness.get_draw(pop_data.index)
        if age_start == age_end:
            age = age_draw * (pop_data.creation_window / pd.Timedelta(days=365)) + age_start
        else:
            age = age_draw * (age_end - age_start) + age_start

        core_population = pd.DataFrame({'entrance_time': pop_data.creation_time,
                                        'age': age.values}, index=pop_data.index)
        self.register(core_population)

        location = self.config.input_data.location
        population = _build_population(core_population, location, self.randomness)
        self.population_view.update(population)


def _build_population(core_population, location, randomness_stream):
    index = core_population.index

    population = pd.DataFrame(
        {'age': core_population['age'],
         'entrance_time': core_population['entrance_time'],
         'sex': randomness_stream.choice(index, ['Male', 'Female'], additional_key='sex_choice'),
         'alive': pd.Series('alive', index=index),
         'location': location,
         'exit_time': pd.NaT, },
        index=index)
    return population


def _non_crn_build_population(index, age_start, age_end, location, creation_time, creation_window, randomness_stream):
    if age_start == age_end:
        age = randomness_stream.get_draw(index) * (creation_window / pd.Timedelta(days=365)) + age_start
    else:
        age = randomness_stream.get_draw(index)*(age_end - age_start) + age_start

    population = pd.DataFrame(
        {'age': age,
         'sex': randomness_stream.choice(index, ['Male', 'Female'], additional_key='sex_choice'),
         'alive': pd.Series('alive', index=index),
         'location': location,
         'entrance_time': creation_time,
         'exit_time': pd.NaT, },
        index=index)
    return population


def build_table(value, year_start, year_end, columns=('age', 'year', 'sex', 'value')):
    value_columns = columns[3:]
    if not isinstance(value, list):
        value = [value]*len(value_columns)

    if len(value) != len(value_columns):
        raise ValueError('Number of values must match number of value columns')

    rows = []
    for age in range(0, 140):
        for year in range(year_start, year_end+1):
            for sex in ['Male', 'Female', 'Both']:
                r_values = []
                for v in value:
                    if v is None:
                        r_values.append(np.random.random())
                    elif callable(v):
                        r_values.append(v(age, sex, year))
                    else:
                        r_values.append(v)
                rows.append([age, year, sex] + r_values)
    return pd.DataFrame(rows, columns=['age', 'year', 'sex'] + list(value_columns))


def make_dummy_column(name, initial_value):
    class DummyColumnMaker:
        def setup(self, builder):
            self.population_view = builder.population.get_view([name])
            builder.population.initializes_simulants(self.make_column,
                                                     creates_columns=[name])

        def make_column(self, pop_data):
            self.population_view.update(pd.Series(initial_value, index=pop_data.index, name=name))

        def __repr__(self):
            return f"dummy_column(name={name}, initial_value={initial_value})"
    return DummyColumnMaker()


def get_randomness(key='test', clock=lambda: pd.Timestamp(1990, 7, 2), seed=12345, for_initialization=False):
    return randomness.RandomnessStream(key, clock, seed=seed, for_initialization=for_initialization)


def reset_mocks(mocks):
    for mock in mocks:
        mock.reset_mock()


def metadata(file_path):
    return {'layer': 'override', 'source': os.path.realpath(file_path)}



import os
import pytest

from vivarium.framework.configuration import build_simulation_configuration, build_model_specification
from vivarium.testing_utilities import metadata


@pytest.fixture(scope='function')
def base_config():
    config = build_simulation_configuration()
    config.update({
        'time': {
            'start': {
                'year': 1990,
            },
            'end': {
                'year': 2010
            },
            'step_size': 30.5
        }
    }, **metadata(__file__))
    return config


@pytest.fixture(scope='function')
def model_specification(mocker):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    user_config = test_dir + '/test_data/mock_user_config.yaml'
    model_spec = test_dir + '/test_data/mock_model_specification.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.os.path.expanduser')
    expand_user_mock.return_value = user_config

    return build_model_specification(model_spec)



import os

import pytest
import yaml

from vivarium.framework.configuration import (ConfigurationError, build_simulation_configuration,
                                              build_model_specification, validate_model_specification_file,
                                              _get_default_specification, DEFAULT_PLUGINS)


def test_get_default_specification_user_config(mocker):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    user_config = test_dir + '/../test_data/mock_user_config.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.os.path.expanduser')
    expand_user_mock.return_value = user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')

    with open(user_config) as f:
        data = {'configuration': yaml.load(f)}

    data.update(DEFAULT_PLUGINS)
    data.update({'components': None})

    assert default_spec.to_dict() == data


def test_get_default_specification_no_user_config(mocker):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    user_config = test_dir + '/../test_data/oh_no_nothing_here.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.os.path.expanduser')
    expand_user_mock.return_value = user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')
    data = {'components': None}
    data.update(DEFAULT_PLUGINS)

    assert default_spec.to_dict() == data


def test_validate_model_specification_failures():
    with pytest.raises(ConfigurationError):
        validate_model_specification_file('made_up_file.yaml')

    with pytest.raises(ConfigurationError):
        test_dir = os.path.dirname(os.path.realpath(__file__))
        model_spec = test_dir + '/../test_data/bad_model_specification.txt'
        assert os.path.exists(model_spec), 'Test directory structure is broken'
        validate_model_specification_file(model_spec)


def test_validate_model_specification():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    model_spec = test_dir + '/../test_data/mock_model_specification.yaml'
    assert os.path.exists(model_spec), 'Test directory structure is broken'
    validate_model_specification_file(model_spec)


def test_build_simulation_configuration(mocker):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    user_config = test_dir + '/../test_data/mock_user_config.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.os.path.expanduser')
    expand_user_mock.return_value = user_config

    config = build_simulation_configuration()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')

    with open(user_config) as f:
        data = yaml.load(f)

    assert config.to_dict() == data


def test_build_model_specification_failure():
    with pytest.raises(ConfigurationError):
        build_model_specification('made_up_file.yaml')

    with pytest.raises(ConfigurationError):
        test_dir = os.path.dirname(os.path.realpath(__file__))
        model_spec = test_dir + '/../test_data/bad_model_specification.txt'
        assert os.path.exists(model_spec), 'Test directory structure is broken'
        build_model_specification(model_spec)


def test_build_model_specification(mocker):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    user_config = test_dir + '/../test_data/mock_user_config.yaml'
    model_spec = test_dir + '/../test_data/mock_model_specification.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.os.path.expanduser')
    expand_user_mock.return_value = user_config

    loaded_model_spec = build_model_specification(model_spec)

    test_data = DEFAULT_PLUGINS

    with open(model_spec) as f:
        model_data = yaml.load(f)

    test_data.update(model_data)

    with open(user_config) as f:
        user_data = yaml.load(f)

    test_data['configuration'].update(user_data)

    assert loaded_model_spec.to_dict() == test_data



"""A set of tools for managing data lookups."""
from numbers import Number
from datetime import datetime, timedelta
import warnings

import pandas as pd

from vivarium.interpolation import Interpolation


class TableView:
    def __call__(self, index):
        raise NotImplementedError()

    def __repr__(self):
        return "TableView()"


class InterpolatedTableView(TableView):
    """A callable that returns the result of an interpolation function over input data.

    Parameters
    ----------
    interpolation : callable
    population_view : `vivarium.framework.population.PopulationView`
    clock : callable

    Notes
    -----
    These cannot be created directly. Use the `lookup` method on the builder during setup.
    """

    def __init__(self, data, population_view, key_columns, parameter_columns, interpolation_order, clock=None):
        if data is None or (isinstance(data, pd.DataFrame) and data.empty):
            raise ValueError("Must supply some data")

        self._data = data
        self._interpolation = None
        self.population_view = population_view
        self._key_columns = key_columns
        self._parameter_columns = parameter_columns
        self._interpolation_order = interpolation_order
        self.clock = clock

    @property
    def interpolation(self):
        if self._interpolation is None:
            data = self._data
            if callable(data) and not isinstance(data, Interpolation):
                data = data()

            if isinstance(data, Interpolation):
                self._interpolation = data
                warnings.warn("Creating lookup tables from pre-initialized Intrepolation objects is deprecated. "
                              "Use key_columns and parameter_columns to control interpolation. If that isn't possible "
                              "then please raise an issue with your use case.", DeprecationWarning)
            else:
                self._interpolation = Interpolation(data, self._key_columns, self._parameter_columns,
                                                    order=self._interpolation_order)
        return self._interpolation

    def __call__(self, index):
        pop = self.population_view.get(index)

        if self.clock:
            current_time = self.clock()
            fractional_year = current_time.year
            fractional_year += current_time.timetuple().tm_yday / 365.25
            pop['year'] = fractional_year

        return self.interpolation(pop)

    def __repr__(self):
        return "InterpolatedTableView()"


class ScalarView(TableView):
    def __init__(self, value):
        self.value = value

    def __call__(self, index):
        return pd.Series(self.value, index=index)

    def __repr__(self):
        return "ScalarView(value={})".format(self.value)


class InterpolatedDataManager:
    """Container for interpolation functions over input data. Interpolation can
    be turned off on a case by case basis in which case the data will be
    delegated to MergedTableManager instead.

    Notes
    -----
    Client code should never access this class directly. Use ``lookup`` on the builder during setup
    to get references to TableView objects.
    """

    configuration_defaults = {
        'interpolation': {
            'order': 1,
        }
    }

    def setup(self, builder):
        self._pop_view_builder = builder.population.get_view
        self.clock = builder.time.clock()
        self._interpolation_order = builder.configuration.interpolation.order
        if self._interpolation_order not in [0, 1]:
            raise ValueError('Only order 0 and order 1 interpolations are supported. '
                             f'You specified {self._interpolation_order}')

    def build_table(self, data, key_columns, parameter_columns):
        """Construct a TableView from a ``pandas.DataFrame``. An interpolation
        function of the specified order will be calculated for each permutation
        of the set of key_columns. The columns in parameter_columns will be used
        as parameters for the interpolation functions which will estimate all
        remaining columns in the table.

        If parameter_columns is empty then no interpolation will be
        attempted and the data will be delegated to MergedTableManager.build_table.


        Parameters
        ----------
        data        : pandas.DataFrame
                      The source data which will be accessible through the resulting TableView.
        key_columns : [str]
                      Columns used to select between interpolation functions. These
                      should be the non-continuous variables in the data. For
                      example 'sex' in data about a population.
        parameter_columns : [str]
                      The columns which contain the parameters to the interpolation functions.
                      These should be the continuous variables. For example 'age'
                      in data about a population.

        Returns
        -------
        TableView
        """

        # Note datetime catches pandas timestamps
        if isinstance(data, Number) or isinstance(data, datetime) or isinstance(data, timedelta):
            return ScalarView(data)

        view_columns = sorted((set(key_columns) | set(parameter_columns)) - {'year'})
        return InterpolatedTableView(data, self._pop_view_builder(view_columns),
                                     key_columns, parameter_columns, self._interpolation_order,
                                     self.clock if 'year' in parameter_columns else None)

    def __repr__(self):
        return "InterpolatedDataManager()"


class LookupTableInterface:

    def __init__(self, manager):
        self._lookup_table_manager = manager

    def build_table(self, data, key_columns=('sex',), parameter_columns=('age', 'year',)):
        return self._lookup_table_manager.build_table(data, key_columns, parameter_columns)



import pytest
import pandas as pd
import numpy as np

from vivarium.framework.values import (set_combiner, list_combiner,
                                       joint_value_post_processor, ValuesManager)


@pytest.fixture
def manager(mocker):
    manager = ValuesManager()
    builder = mocker.MagicMock()
    builder.step_size = lambda: lambda: pd.Timedelta(days=1)
    manager.setup(builder)
    return manager


def test_replace_combiner(manager):
    value = manager.register_value_producer('test', source=lambda: 1)

    assert value() == 1

    manager.register_value_modifier('test', modifier=lambda v: 42)
    assert value() == 42

    manager.register_value_modifier('test', lambda v: 84)
    assert value() == 84


def test_joint_value(manager):
    # This is the normal configuration for PAF and disability weight type values
    index = pd.Index(range(10))

    value = manager.register_value_producer('test',
                                            source=lambda idx: [pd.Series(0, index=idx)],
                                            preferred_combiner=list_combiner,
                                            preferred_post_processor=joint_value_post_processor)
    assert np.all(value(index) == 0)

    manager.register_value_modifier('test', modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.5)

    manager.register_value_modifier('test', modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.75)


def test_set_combiner(manager):
    # This is the normal configuration for collecting lists of meids for calculating cause deleted tables
    value = manager.register_value_producer('test', source=lambda: set(), preferred_combiner=set_combiner)

    assert value() == set()

    manager.register_value_modifier('test', modifier=lambda: 'thing one')
    assert value() == {'thing one'}

    manager.register_value_modifier('test', modifier=lambda: 'thing one')
    assert value() == {'thing one'}  # duplicates are truly removed

    manager.register_value_modifier('test', modifier=lambda: 'thing two')
    assert value() == {'thing one', 'thing two'}  # but unique values are collected


def test_contains(manager):
    value = 'test_value'
    rate = 'test_rate'

    assert value not in manager
    assert rate not in manager

    manager.register_value_producer('test_value', source=lambda: 1)
    assert value in manager
    assert rate not in manager

    manager.register_rate_producer('test_rate', source=lambda: 1)
    assert value in manager
    assert rate in manager



import pandas as pd
import numpy as np

from vivarium.framework.randomness import choice
from vivarium.framework.state_machine import Machine, State, Transition

from vivarium.interface.interactive import setup_simulation


def _population_fixture(column, initial_value):
    class PopFixture:
        def setup(self, builder):
            self.population_view = builder.population.get_view([column])
            builder.population.initializes_simulants(self.inner, creates_columns=[column])

        def inner(self, pop_data):
            self.population_view.update(pd.Series(initial_value, index=pop_data.index))
    return PopFixture()


def _even_population_fixture(column, values):
    class pop_fixture:
        def setup(self, builder):
            self.population_view = builder.population.get_view([column])
            builder.population.initializes_simulants(self.inner, creates_columns=[column])

        def inner(self, pop_data):
            self.population_view.update(choice('start', pop_data.index, values))

    return pop_fixture()


def test_transition():
    done_state = State('done')
    start_state = State('start')
    start_state.add_transition(done_state)
    machine = Machine('state', states=[start_state, done_state])

    simulation = setup_simulation([machine, _population_fixture('state', 'start')])
    event_time = simulation.clock.time + simulation.clock.step_size
    machine.transition(simulation.population.population.index, event_time)
    assert np.all(simulation.population.population.state == 'done')


def test_choice(base_config):
    base_config.population.update({'population_size': 10000})
    a_state = State('a')
    b_state = State('b')
    start_state = State('start')
    start_state.add_transition(a_state, probability_func=lambda agents: np.full(len(agents), 0.5))
    start_state.add_transition(b_state, probability_func=lambda agents: np.full(len(agents), 0.5))
    machine = Machine('state', states=[start_state, a_state, b_state])

    simulation = setup_simulation([machine, _population_fixture('state', 'start')], base_config)
    event_time = simulation.clock.time + simulation.clock.step_size
    machine.transition(simulation.population.population.index, event_time)
    a_count = (simulation.population.population.state == 'a').sum()
    assert round(a_count/len(simulation.population.population), 1) == 0.5


def test_null_transition(base_config):
    base_config.population.update({'population_size': 10000})
    a_state = State('a')
    start_state = State('start')
    start_state.add_transition(a_state, probability_func=lambda agents: np.full(len(agents), 0.5))
    start_state.allow_self_transitions()

    machine = Machine('state', states=[start_state, a_state])

    simulation = setup_simulation([machine, _population_fixture('state', 'start')], base_config)
    event_time = simulation.clock.time + simulation.clock.step_size
    machine.transition(simulation.population.population.index, event_time)
    a_count = (simulation.population.population.state == 'a').sum()
    assert round(a_count/len(simulation.population.population), 1) == 0.5


def test_no_null_transition(base_config):
    base_config.population.update({'population_size': 10000})
    a_state = State('a')
    b_state = State('b')
    start_state = State('start')
    a_transition = Transition(start_state, a_state)
    b_transition = Transition(start_state, b_state)
    start_state.transition_set.allow_null_transition = False
    start_state.transition_set.extend((a_transition, b_transition))
    machine = Machine('state')
    machine.states.extend([start_state, a_state, b_state])

    simulation = setup_simulation([machine, _population_fixture('state', 'start')], base_config)
    event_time = simulation.clock.time + simulation.clock.step_size
    machine.transition(simulation.population.population.index, event_time)
    a_count = (simulation.population.population.state == 'a').sum()
    assert round(a_count/len(simulation.population.population), 1) == 0.5


def test_side_effects():
    class DoneState(State):
        def setup(self, builder):
            super().setup(builder)
            self.population_view = builder.population.get_view(['count'])

        def _transition_side_effect(self, index, event_time):
            pop = self.population_view.get(index)
            self.population_view.update(pop['count'] + 1)

    done_state = DoneState('done')
    start_state = State('start')
    start_state.add_transition(done_state)
    done_state.add_transition(start_state)

    machine = Machine('state', states=[start_state, done_state])

    simulation = setup_simulation([machine, _population_fixture('state', 'start'), _population_fixture('count', 0)])
    event_time = simulation.clock.time + simulation.clock.step_size
    machine.transition(simulation.population.population.index, event_time)
    assert np.all(simulation.population.population['count'] == 1)
    machine.transition(simulation.population.population.index, event_time)
    assert np.all(simulation.population.population['count'] == 1)
    machine.transition(simulation.population.population.index, event_time)
    assert np.all(simulation.population.population['count'] == 2)



import os

import pytest

from vivarium.framework.configuration import build_simulation_configuration
from vivarium.framework.components.manager import (ComponentManager, ComponentConfigError,
                                                   _setup_components, _apply_component_default_configuration)

from .mocks import MockComponentA, MockComponentB


@pytest.fixture
def apply_default_config_mock(mocker):
    return mocker.patch('vivarium.framework.components.manager._apply_component_default_configuration')


def test__apply_component_default_configuration():

    class UnladenSwallow:

        configuration_defaults = {
            'unladen_swallow': {
                'airspeed_velocity': 11,
            }
        }

    us = UnladenSwallow()
    config = build_simulation_configuration()
    assert 'unladen_swallow' not in config
    _apply_component_default_configuration(config, us)
    assert config.unladen_swallow.metadata('airspeed_velocity') == [
        {'layer': 'component_configs', 'value': 11,
         'source': os.path.realpath(__file__), 'default': False}
    ]

    us = UnladenSwallow()
    us.__module__ = '__main__'
    config = build_simulation_configuration()
    assert 'unladen_swallow' not in config
    _apply_component_default_configuration(config, us)
    assert config.unladen_swallow.metadata('airspeed_velocity') == [
        {'layer': 'component_configs', 'value': 11, 'source': '__main__', 'default': False}
    ]


def test__setup_components(mocker, apply_default_config_mock):
    config = build_simulation_configuration()
    builder = mocker.Mock()

    components = [None, MockComponentA('Eric'),  MockComponentB('half', 'a', 'bee')]

    with pytest.raises(ComponentConfigError):
        _setup_components(builder, components, config)

    duplicate = MockComponentA('Eric')
    components = [duplicate, duplicate]
    finished = _setup_components(builder, components, config)

    apply_default_config_mock.assert_called_once_with(config, duplicate)
    assert [duplicate] == finished
    apply_default_config_mock.reset_mock()

    components = [MockComponentA('Eric'), MockComponentB('half', 'a', 'bee')]
    finished = _setup_components(builder, components, config)
    mock_a, mock_b = finished

    assert apply_default_config_mock.mock_calls == [mocker.call(config, component) for component in finished]

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder
    assert mock_b.args == ('half', 'a', 'bee')


def test_ComponentManager_add_components():
    manager = ComponentManager()

    components = [None, MockComponentA('Eric'), MockComponentB('half', 'a', 'bee')]
    for list_type in ['_managers', '_components']:
        manager._add_components(getattr(manager, list_type), components)
        assert getattr(manager, list_type) == components
        setattr(manager, list_type, [])

    components.append(components[:])
    for list_type in ['_managers', '_components']:
        manager._add_components(getattr(manager, list_type), components)
        assert getattr(manager, list_type) == 2*components[:-1]


def test_ComponentManager__setup_components(mocker):
    config = build_simulation_configuration()
    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager

    manager.add_components([None, MockComponentA('Eric'),
                            MockComponentB('half', 'a', 'bee')])
    with pytest.raises(ComponentConfigError):
        manager.setup_components(builder, config)

    manager._components = []
    manager.add_components([MockComponentA('Eric'), MockComponentB('half', 'a', 'bee')])
    manager.setup_components(builder, config)

    mock_a, mock_b, mock_b_child1, mock_b_child2, mock_b_child3 = manager._components

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder
    assert mock_b_child1.args == ('half',)
    assert mock_b_child1.builder_used_for_setup is builder
    assert mock_b_child2.args == ('a',)
    assert mock_b_child2.builder_used_for_setup is builder
    assert mock_b_child3.args == ('bee',)
    assert mock_b_child3.builder_used_for_setup is builder



"""A simple component to allow users to store outputs in the simulation."""


class Metrics:
    """This class declares a value pipeline that allows other components to store summary metrics."""
    def setup(self, builder):
        self.metrics = builder.value.register_value_producer('metrics', source=lambda index: {})



import pandas as pd

from vivarium.framework.state_machine import State, Machine, Transition
from vivarium.framework.utilities import rate_to_probability
from vivarium.framework.values import list_combiner, joint_value_post_processor


class DiseaseTransition(Transition):

    def __init__(self, name, rate, input_state, output_state, **kwargs):
        super().__init__(input_state, output_state, probability_func=self._probability, **kwargs)
        self.name = name
        self.base_rate = lambda index: pd.Series(rate, index=index)

    def setup(self, builder):
        self.transition_rate = builder.value.register_rate_producer(f'{self.name}_rate',
                                                                      source=self._risk_deleted_rate)
        self.joint_population_attributable_fraction = builder.value.register_value_producer(
            f'{self.name}_rate.population_attributable_fraction',
            source=lambda index: [pd.Series(0, index=index)],
            preferred_combiner=list_combiner,
            preferred_post_processor=joint_value_post_processor)

    def _probability(self, index):
        effective_rate = self.transition_rate(index)
        return rate_to_probability(effective_rate)

    def _risk_deleted_rate(self, index):
        return self.base_rate(index) * (1 - self.joint_population_attributable_fraction(index))


class DiseaseState(State):

    def __init__(self, state_name, excess_mortality_rate=0, **kwargs):
        super().__init__(state_name, **kwargs)
        self._excess_mortality_rate = excess_mortality_rate

    def setup(self, builder):
        """Performs this component's simulation setup.

        Parameters
        ----------
        builder : `engine.Builder`
            Interface to several simulation tools.
        """
        super().setup(builder)
        self.clock = builder.time.clock()

        self.excess_mortality_rate = builder.value.register_rate_producer(
            f'{self.state_id}.excess_mortality_rate',
            source=self.risk_deleted_excess_mortality_rate
        )
        self.excess_mortality_rate_paf = builder.value.register_value_producer(
            f'{self.state_id}.excess_mortality_rate.population_attributable_fraction',
            source=lambda index: [pd.Series(0, index=index)],
            preferred_combiner=list_combiner,
            preferred_post_processor=joint_value_post_processor
        )

        builder.value.register_value_modifier('mortality_rate', self.add_in_excess_mortality)
        self.population_view = builder.population.get_view(
            [self._model], query=f"alive == 'alive' and {self._model} == '{self.state_id}'")

    def add_transition(self, transition_name, output, rate=1e6, **kwargs):
        t = DiseaseTransition(transition_name, rate, self, output, **kwargs)
        self.transition_set.append(t)
        return t

    def risk_deleted_excess_mortality_rate(self, index):
        return pd.Series(self._excess_mortality_rate, index=index) * (1 - self.excess_mortality_rate_paf(index))

    def add_in_excess_mortality(self, index, mortality_rates):
        affected = self.population_view.get(index)
        mortality_rates.loc[affected.index] += self.excess_mortality_rate(affected.index)

        return mortality_rates


class DiseaseModel(Machine):

    def __init__(self, disease, initial_state, cause_specific_mortality_rate=0., **kwargs):
        super().__init__(disease, **kwargs)
        self.initial_state = initial_state.state_id
        self._cause_specific_mortality_rate = cause_specific_mortality_rate

    def setup(self, builder):
        super().setup(builder)
        self.cause_specific_mortality_rate = builder.value.register_rate_producer(
            f'{self.state_column}.cause_specific_mortality_rate',
            source=lambda index: pd.Series(self._cause_specific_mortality_rate, index=index)
        )
        builder.value.register_value_modifier('mortality_rate', modifier=self.delete_cause_specific_mortality)
        builder.value.register_value_modifier('metrics', modifier=self.metrics)

        creates_columns = [self.state_column]
        builder.population.initializes_simulants(self.on_initialize_simulants, creates_columns=creates_columns)
        self.population_view = builder.population.get_view(['age', 'sex', self.state_column])

        builder.event.register_listener('time_step', self.on_time_step)

    def on_initialize_simulants(self, pop_data):
        condition_column = pd.Series(self.initial_state, index=pop_data.index, name=self.state_column)
        self.population_view.update(condition_column)

    def on_time_step(self, event):
        self.transition(event.index, event.time)

    def delete_cause_specific_mortality(self, index, rates):
        return rates - self.cause_specific_mortality_rate(index)

    def metrics(self, index, metrics):
        pop = self.population_view.get(index, query="alive == 'alive'")
        metrics[self.state_column + '_prevalent_cases'] = len(pop[pop[self.state_column] != self.initial_state])
        return metrics


class SIS_DiseaseModel:

    configuration_defaults = {
        'disease': {
            'incidence': 0.005,
            'remission': 0.05,
            'excess_mortality': 0.01,
        }
    }

    def __init__(self, disease_name):
        self.name = disease_name
        self.configuration_defaults = {disease_name: SIS_DiseaseModel.configuration_defaults['disease']}

    def setup(self, builder):
        config = builder.configuration[self.name]

        susceptible_state = DiseaseState(f'susceptible_to_{self.name}')
        infected_state = DiseaseState(f'infected_with_{self.name}',
                                      excess_mortality_rate=config.excess_mortality)

        susceptible_state.allow_self_transitions()
        susceptible_state.add_transition(f'{infected_state.state_id}.incidence', infected_state, rate=config.incidence)
        infected_state.allow_self_transitions()
        infected_state.add_transition(f'{infected_state.state_id}.remission', susceptible_state, rate=config.remission)

        # Reasonable approximation for short duration diseases.
        case_fatality_rate = config.excess_mortality / (config.excess_mortality + config.remission)
        cause_specific_mortality = config.incidence * case_fatality_rate

        model = DiseaseModel(self.name,
                             initial_state=susceptible_state,
                             cause_specific_mortality_rate=cause_specific_mortality,
                             states=[susceptible_state, infected_state])
        builder.components.add_components([model])



import pandas as pd
import numpy as np

from vivarium.framework.event import Event, EventManager


def test_split_event():
    index1 = pd.Index(range(10))
    index2 = pd.Index(range(5))

    e1 = Event(index1)
    e2 = e1.split(index2)

    assert e1.index is index1
    assert e2.index is index2


def test_emission():
    signal = [False]

    def listener(*_, **__):
        signal[0] = True

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, unit='D')
    emitter = manager.get_emitter('test_event')
    manager.register_listener('test_event', listener)
    emitter(Event(None))

    assert signal[0]

    signal[0] = False

    emitter = manager.get_emitter('test_unheard_event')
    emitter(Event(None))
    assert not signal[0]


def test_listener_priority():
    signal = [False, False, False]

    def listener1(*_, **__):
        signal[0] = True
        assert not signal[1]
        assert not signal[2]

    def listener2(*_, **__):
        signal[1] = True
        assert signal[0]
        assert not signal[2]

    def listener3(*_, **__):
        signal[2] = True
        assert signal[0]
        assert signal[1]

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, 'D')
    emitter = manager.get_emitter('test_event')
    manager.register_listener('test_event', listener1, priority=0)
    manager.register_listener('test_event', listener2)
    manager.register_listener('test_event', listener3, priority=9)

    emitter(Event(None))
    assert np.all(signal)


def test_contains():
    event = 'test_event'

    manager = EventManager()
    assert event not in manager
    manager.get_emitter(event)
    assert event in manager




class MockComponentA:
    def __init__(self, *args):
        self.args = args
        self.builder_used_for_setup = None


class MockComponentB(MockComponentA):
    def setup(self, builder):
        self.builder_used_for_setup = builder

        if len(self.args) > 1:
            children = []
            for arg in self.args:
                children.append(MockComponentB(arg))
            builder.components.add_components(children)
        builder.value.register_value_modifier('metrics', self.metrics)

    def metrics(self, _, metrics):
        if 'test' in metrics:
            metrics['test'] += 1
        else:
            metrics['test'] = 1
        return metrics


class Listener(MockComponentB):
    def __init__(self, *args):
        super().__init__(*args)
        self.post_setup_called = False
        self.time_step__prepare_called = False
        self.time_step_called = False
        self.time_step__cleanup_called = False
        self.collect_metrics_called = False
        self.simulation_end_called = False

    def setup(self, builder):
        super().setup(builder)
        builder.event.register_listener('post_setup', self.on_post_setup)
        builder.event.register_listener('time_step__prepare', self.on_time_step__prepare)
        builder.event.register_listener('time_step', self.on_time_step)
        builder.event.register_listener('time_step__cleanup', self.on_time_step__cleanup)
        builder.event.register_listener('collect_metrics', self.on_collect_metrics)
        builder.event.register_listener('simulation_end', self.on_simulation_end)

    def on_post_setup(self, _):
        self.post_setup_called = True

    def on_time_step__prepare(self, _):
        self.time_step__prepare_called = True

    def on_time_step(self, _):
        self.time_step_called = True

    def on_time_step__cleanup(self, _):
        self.time_step__cleanup_called = True

    def on_collect_metrics(self, _):
        self.collect_metrics_called = True

    def on_simulation_end(self, _):
        self.simulation_end_called = True



from math import ceil

from vivarium.framework.configuration import build_simulation_configuration, build_model_specification
from vivarium.framework.plugins import PluginManager
from vivarium.framework.engine import SimulationContext

from .utilities import run_from_ipython, log_progress, raise_if_not_setup


class InteractiveContext(SimulationContext):

    def __init__(self, configuration, components, plugin_manager=None):
        super().__init__(configuration, components, plugin_manager)
        self._initial_population = None
        self._setup = False

    def setup(self):
        super().setup()
        self._start_time = self.clock.time
        self.initialize_simulants()
        self._setup = True

    def initialize_simulants(self):
        super().initialize_simulants()
        self._initial_population = self.population.population

    def reset(self):
        # This is super crude, but should work for a great deal of components.
        self.population._population = self._initial_population
        self.clock._time = self._start_time

    @raise_if_not_setup(system_type='run')
    def run(self, with_logging=True):
        return self.run_until(self.clock.stop_time, with_logging=with_logging)

    @raise_if_not_setup(system_type='run')
    def run_for(self, duration, with_logging=True):
        return self.run_until(self.clock.time + duration, with_logging=with_logging)

    @raise_if_not_setup(system_type='run')
    def run_until(self, end_time, with_logging=True):
        if not isinstance(end_time, type(self.clock.time)):
            raise ValueError(f"Provided time must be an instance of {type(self.clock.time)}")

        iterations = int(ceil((end_time - self.clock.time)/self.clock.step_size))
        self.take_steps(number_of_steps=iterations, with_logging=with_logging)
        assert self.clock.time - self.clock.step_size < end_time <= self.clock.time
        return iterations

    @raise_if_not_setup(system_type='run')
    def step(self, step_size=None):  # TODO: consider renaming to take_step for similarity with sim.take_steps
        old_step_size = self.clock.step_size
        if step_size is not None:
            if not isinstance(step_size, type(self.clock.step_size)):
                raise ValueError(f"Provided time must be an instance of {type(self.clock.step_size)}")
            self.clock._step_size = step_size
        super().step()
        self.clock._step_size = old_step_size

    @raise_if_not_setup(system_type='run')
    def take_steps(self, number_of_steps=1, step_size=None, with_logging=True):
        if not isinstance(number_of_steps, int):
            raise ValueError('Number of steps must be an integer.')

        if run_from_ipython() and with_logging:
            for _ in log_progress(range(number_of_steps), name='Step'):
                self.step(step_size)
        else:
            for _ in range(number_of_steps):
                self.step(step_size)

    @raise_if_not_setup(system_type='value')
    def list_values(self):
        return list(self.values.keys())

    @raise_if_not_setup(system_type='value')
    def get_values(self):
        return self.values.items()

    @raise_if_not_setup(system_type='value')
    def get_value(self, value_pipeline_name):
        return self.values.get_value(value_pipeline_name)

    @raise_if_not_setup(system_type='event')
    def list_events(self):
        return self.events.list_events()

    @raise_if_not_setup(system_type='event')
    def get_listeners(self, event_name):
        if event_name not in self.events:
            raise ValueError(f'No event {event_name} in system.')
        return self.events.get_listeners(event_name)

    @raise_if_not_setup(system_type='event')
    def get_emitter(self, event_name):
        if event_name not in self.events:
            raise ValueError(f'No event {event_name} in system.')
        return self.events.get_emitter(event_name)

    @raise_if_not_setup(system_type='component')
    def get_components(self):
        return [component for component in self.component_manager._components + self.component_manager._managers]

    @raise_if_not_setup(system_type='component')
    def reload_component(self, component):
        raise NotImplementedError()

    @raise_if_not_setup(system_type='component')
    def replace_component(self, old_component, new_component):
        self.component_manager._components.remove(old_component)
        new_component.setup(self.builder)
        self.component_manager.add_components([new_component])


def initialize_simulation(components, input_config=None, plugin_config=None):
    config = build_simulation_configuration()
    config.update(input_config)
    plugin_manager = PluginManager(plugin_config)

    return InteractiveContext(config, components, plugin_manager)


def setup_simulation(components, input_config=None, plugin_config=None):
    simulation = initialize_simulation(components, input_config, plugin_config)
    simulation.setup()

    return simulation


def initialize_simulation_from_model_specification(model_specification_file):
    model_specification = build_model_specification(model_specification_file)

    plugin_config = model_specification.plugins
    component_config = model_specification.components
    simulation_config = model_specification.configuration

    plugin_manager = PluginManager(plugin_config)
    component_config_parser = plugin_manager.get_plugin('component_configuration_parser')
    components = component_config_parser.get_components(component_config)

    return InteractiveContext(simulation_config, components, plugin_manager)


def setup_simulation_from_model_specification(model_specification_file):
    simulation = initialize_simulation_from_model_specification(model_specification_file)
    simulation.setup()

    return simulation



"""Defines the parsing rules for the ``components`` section of ``yaml`` configuration files."""
from typing import Sequence, Tuple, List, Dict, Union


from vivarium.config_tree import ConfigTree
from vivarium.framework.utilities import import_by_path
from .manager import ComponentConfigError


class ParsingError(ComponentConfigError):
    """Error raised when component configurations are not specified correctly."""
    pass


class ComponentConfigurationParser:
    """Parses component configuration and initializes components.

    To define your own set of parsing rules, you should write a parser class that
    inherits from this class and overrides ``parse_component_config``.  You can then
    define a set of parsing rules that turn component configuration information
    into a list of strings where each string is the full python import path to the class
    followed by a set of parentheses containing initialization arguments.

    For example, say your component configuration specifies ``FancyClass`` from the
    ``important_module`` of your ``made_up_package`` and that ``FancyClass`` has two
    initialization parameters, ``'important_thing1'`` and ``'important_thing2'``.
    Your implementation of ``parse_component_config`` needs to generate the string
    ``'made_up_package.important_module.FancyClass("important_thing1", "important_thing2")'``
    and include it in the list of components to generate.

    Currently, all classes that are initialized from the ``yaml`` configuration must either
    take no arguments or take arguments specified as strings.
    """

    def get_components(self, component_config: Union[ConfigTree, List]) -> List:
        """Extracts component specifications from configuration information and returns initialized components.

        Parameters
        ----------
        component_config :
            A hierarchical component specification blob. This configuration information needs to be parsable
            into a full import path and a set of initialization arguments by the ``parse_component_config``
            method.

        Returns
        -------
        List
            A list of initialized components.
        """
        if isinstance(component_config, ConfigTree):
            component_list = self.parse_component_config(component_config.to_dict())
        else:  # Components were specified in a list rather than a tree.
            component_list = component_config
        component_list = _prep_components(component_list)
        return _import_and_instantiate_components(component_list)

    def parse_component_config(self, component_config: Dict[str, Union[Dict, List]]) -> List[str]:
        """Parses a hierarchical component specification into a list of standardized component definitions.

        This default parser expects component configurations as a list of dicts. Each dict at the top level
        corresponds to a different package and has a single key. This key may be just the name of the package
        or a Python style import path to the module in which components live. The values of the top level dicts
        are a list of dicts or strings. If dicts, the keys are another step along the import path. If strings,
        the strings are representations of calls to the class constructor of components to be generated. This
        pattern may be arbitrarily nested.

        Parameters
        ----------
        component_config :
            A hierarchical component specification blob.

        Returns
        -------
        List
            A list of standardized component definitions. Component definition strings are specified as
            ``'absolute.import.path.ClassName("argument1", "argument2", ...)'``.

        """
        return _parse_component_config(component_config)


def _parse_component_config(component_config: Union[List[str], Dict[str, Union[Dict, List]]]) -> List[str]:

    def _process_level(level, prefix):
        if isinstance(level, list):
            return ['.'.join(prefix + [child]) for child in level]

        component_list = []
        for name, child in level.items():
            component_list.extend(_process_level(child, prefix + [name]))

        return component_list

    return _process_level(component_config, [])


def _prep_components(component_list: Sequence[str]) -> List[Tuple[str, Tuple[str]]]:
    """Transform component description strings into tuples of component paths and required arguments.

    Parameters
    ----------
    component_list :
        The component descriptions to transform.

    Returns
    -------
    List of component/argument tuples.
    """
    components = []
    for c in component_list:
        path, args_plus = c.split('(')
        cleaned_args = _clean_args(args_plus[:-1].split(','), path)
        components.append((path, cleaned_args))
    return components


def _clean_args(args, path):
    out = []
    for a in args:
        a = a.strip()
        if not a:
            continue

        not_a_valid_string = len(a) < 3 or not ((a[0] == a[-1] == "'") or (a[0] == a[-1] == '"'))
        if not_a_valid_string:
            raise ParsingError(f"Invalid component argument {a} for component {path}")

        out.append(a[1:-1])
    return tuple(out)


def _import_and_instantiate_components(component_list: List[Tuple[str, Tuple[str]]]):
    return [import_by_path(component[0])(*component[1]) for component in component_list]



import numpy as np
import pandas as pd


class BasePopulation:
    configuration_defaults = {
        'population': {
            'age_start': 0,
            'age_end': 100,
        },
    }

    def setup(self, builder):
        self.config = builder.configuration

        self.with_common_random_numbers = bool(self.config.randomness.key_columns)
        self.randomness = builder.randomness.get_stream('age_initialization',
                                                        for_initialization=self.with_common_random_numbers)
        self.register = builder.randomness.register_simulants

        columns_created = ['age', 'sex', 'alive', 'entrance_time']
        builder.population.initializes_simulants(self.generate_population, creates_columns=columns_created)
        self.population_view = builder.population.get_view(columns_created)
        builder.event.register_listener('time_step', self.age_simulants)

    def generate_population(self, pop_data):
        age_start = pop_data.user_data.get('age_start', self.config.population.age_start)
        age_end = pop_data.user_data.get('age_end', self.config.population.age_end)

        age_draw = self.randomness.get_draw(pop_data.index)
        age_window = pop_data.creation_window / pd.Timedelta(days=365) if age_start == age_end else age_end - age_start
        age = age_start + age_draw * age_window

        if self.with_common_random_numbers:
            population = pd.DataFrame({'entrance_time': pop_data.creation_time,
                                       'age': age.values}, index=pop_data.index)
            self.register(population)
            population['sex'] = self.randomness.choice(pop_data.index, ['Male', 'Female'], additional_key='sex_choice')
            population['alive'] = 'alive'
        else:
            population = pd.DataFrame(
                {'age': age.values,
                 'sex': self.randomness.choice(pop_data.index, ['Male', 'Female'], additional_key='sex_choice'),
                 'alive': pd.Series('alive', index=pop_data.index),
                 'entrance_time': pop_data.creation_time},
                index=pop_data.index)

        self.population_view.update(population)

    def age_simulants(self, event):
        population = self.population_view.get(event.index, query="alive == 'alive'")
        population['age'] += event.step_size / pd.Timedelta(days=365)
        self.population_view.update(population)


class Mortality:

    configuration_defaults = {
        'mortality': {
            'mortality_rate': 0.01,
            'life_expectancy': 80,
        }
    }

    def setup(self, builder):
        self.config = builder.configuration.mortality
        self.population_view = builder.population.get_view(['alive'], query="alive == 'alive'")
        self.randomness = builder.randomness.get_stream('mortality')

        self.mortality_rate = builder.value.register_rate_producer('mortality_rate', source=self.base_mortality_rate)

        builder.event.register_listener('time_step', self.determine_deaths)

    def base_mortality_rate(self, index):
        return pd.Series(self.config.mortality_rate, index=index)

    def determine_deaths(self, event):
        effective_rate = self.mortality_rate(event.index)
        effective_probability = 1 - np.exp(-effective_rate)
        draw = self.randomness.get_draw(event.index)
        affected_simulants = draw < effective_probability
        self.population_view.update(pd.Series('dead', index=event.index[affected_simulants]))


class Observer:

    def setup(self, builder):
        self.life_expectancy = builder.configuration.mortality.life_expectancy
        self.population_view = builder.population.get_view(['age', 'alive'])

        builder.value.register_value_modifier('metrics', self.metrics)

    def metrics(self, index, metrics):

        pop = self.population_view.get(index)
        metrics['total_population_alive'] = len(pop[pop.alive == 'alive'])
        metrics['total_population_dead'] = len(pop[pop.alive == 'dead'])

        metrics['years_of_life_lost'] = (self.life_expectancy - pop.age[pop.alive == 'dead']).sum()

        return metrics



from bdb import BdbQuit
import os

import click

from vivarium.framework.engine import run_simulation

from .utilities import verify_yaml

import logging
logging.getLogger(__name__)


@click.group()
def simulate():
    pass


@simulate.command()
@click.argument('model_specification', callback=verify_yaml,
                type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@click.option('--results_directory', '-o', type=click.Path(resolve_path=True),
              default=os.path.expanduser('~/vivarium_results/'),
              help='Top level output directory to write results to')
@click.option('--verbose', '-v', is_flag=True, help='Report each time step')
@click.option('--log', type=click.Path(dir_okay=False, resolve_path=True), help='Path to log file')
@click.option('--pdb', 'with_debugger', is_flag=True, help='Drop into python debugger if an error occurs.')
def run(model_specification, results_directory, verbose, log, with_debugger):
    log_level = logging.DEBUG if verbose else logging.ERROR
    logging.basicConfig(filename=log, level=log_level)

    try:
        run_simulation(model_specification, results_directory)
    except (BdbQuit, KeyboardInterrupt):
        raise
    except Exception as e:
        if with_debugger:
            import pdb
            import traceback
            traceback.print_exc()
            pdb.post_mortem()
        else:
            logging.exception("Uncaught exception {}".format(e))
            raise



"""Provides a class for consistently managing and writing vivarium outputs and output paths."""
import shutil
from collections import defaultdict
import os
from datetime import datetime

import yaml


class ResultsWriter:
    """Writes output files for vivarium simulations.

    Attributes
    ----------
    results_root: str
        The root directory to which results will be written.
    """

    def __init__(self, results_root):
        """
        Parameters
        ----------
        results_root: str
            The root directory to which results will be written.
        """
        self.results_root = results_root
        os.makedirs(results_root, exist_ok=True)
        self._directories = defaultdict(lambda: self.results_root)

    def add_sub_directory(self, key, path):
        """Adds a sub-directory to the results directory.

        Parameters
        ----------
        key: str
            A look-up key for the directory path.
        path: str
            The relative path from the root of the results directory to the sub-directory.

        Returns
        -------
        str:
            The absolute path to the sub-directory.
        """
        sub_dir_path = os.path.join(self.results_root, path)
        os.makedirs(sub_dir_path, exist_ok=True)
        self._directories[key] = sub_dir_path
        return sub_dir_path

    def write_output(self, data, file_name, key=None):
        """Writes output data to disk.

        Parameters
        ----------
        data: pandas.DataFrame or dict
            The data to write to disk.
        file_name: str
            The name of the file to write.
        key: str, optional
            The lookup key for the sub_directory to write results to, if any.
        """
        path = os.path.join(self._directories[key], file_name)
        extension = file_name.split('.')[-1]

        if extension == 'yaml':
            with open(path, 'w') as f:
                yaml.dump(data, f)
        elif extension == 'hdf':
            data.to_hdf(path, 'data', format='table')
        else:
            raise NotImplementedError(
                f"Only 'yaml' and 'hdf' file types are supported. You requested {extension}")

    def copy_file(self, src_path, file_name, key=None):
        """Copies a file unmodified to a location inside the ouput directory.

        Parameters
        ----------
        src_path: str
            Path to the src file
        file_name: str
            name of the destination file
        """
        path = os.path.join(self._directories[key], file_name)
        shutil.copyfile(src_path, path)


def get_results_writer(results_directory, model_specification_file):
    launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    config_name = os.path.basename(model_specification_file.rpartition('.')[0])
    results_root = results_directory + f"/{config_name}/{launch_time}"
    return ResultsWriter(results_root)



from .interactive import (build_simulation_configuration, setup_simulation, setup_simulation_from_model_specification,
                          initialize_simulation, initialize_simulation_from_model_specification)



"""The event framework"""
from collections import defaultdict
from typing import Callable


class Event:
    """An Event object represents the context of an event. It is possible to subclass Event
    to add information for more specialized cases like vivarium.framework.population.PopulationEvent

    Attributes
    ----------
    time  : pandas.Timestamp
            The simulation time at which this event was emitted.
    index : pandas.Index
            An index into the population table containing all simulants effected by this event.
    """

    def __init__(self, index, user_data=None):
        self.index = index
        self.user_data = user_data if user_data is not None else {}
        self.time = None
        self.step_size = None

    def split(self, new_index):
        """Create a new event which is a copy of this one but with a new index.
        """
        new_event = Event(new_index, self.user_data)
        new_event.time = self.time
        new_event.step_size = self.step_size
        return new_event

    def __repr__(self):
        return "Event(user_data={}, time={}, step_size={})".format(self.user_data, self.time, self.step_size)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class _EventChannel:
    def __init__(self, manager):
        self.manager = manager

        self.listeners = [[] for _ in range(10)]

    def emit(self, event):
        """Notifies all listeners to this channel that an event has occurred.

        Parameters
        ----------
        event : Event
            The event to be emitted.
        """
        if hasattr(event, 'time'):
            event.step_size = self.manager.step_size()
            event.time = self.manager.clock() + self.manager.step_size()

        for priority_bucket in self.listeners:
            for listener in sorted(priority_bucket, key=lambda x: x.__name__):
                listener(event)
        return event

    def __repr__(self):
        return "_EventChannel(listeners: {})".format([listener for bucket in self.listeners for listener in bucket])


class EventManager:
    """The configuration for the event system.

    Notes
    -----
    Client code should never need to interact with this class
    except through the decorators in this module and the emitter
    function exposed on the builder during the setup phase.
    """

    def __init__(self):
        self._event_types = defaultdict(lambda: _EventChannel(self))

    def setup(self, builder):
        """Performs this components simulation setup.

        Parameters
        ----------
        builder : vivarium.framework.engine.Builder
            Object giving access to core framework functionality.
        """
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()

    def get_emitter(self, name):
        """Get an emitter function for the named event

        Parameters
        ----------
        name : str
               The name of the event

        Returns
        -------
        callable
            A function that accepts an Event object and distributes
            it to all listeners for this event.
        """
        return self._event_types[name].emit

    def register_listener(self, name, listener, priority=5):
        """Registers a new listener to the named event.

        Parameters
        ----------
        name : str
            The name of the event.
        listener : Callable
            The consumer of the named event.
        priority : int
            Number in range(10) used to assign the ordering in which listeners process the event.
        """
        self._event_types[name].listeners[priority].append(listener)

    def get_listeners(self, name):
        channel = self._event_types[name]
        return {priority: listeners for priority, listeners in enumerate(channel.listeners) if listeners}

    def list_events(self):
        """List all event names known to the event system

        Returns
        -------
        List[Event]
            A list of all known event names

        Notes
        -----
        This value can change after setup if components dynamically create new event labels.
        """
        return list(self._event_types.keys())

    def __contains__(self, item):
        return item in self._event_types

    def __repr__(self):
        return "EventManager(event_types: {})".format(self._event_types.keys())


class EventInterface:

    def __init__(self, event_manager: EventManager):
        self._event_manager = event_manager

    def get_emitter(self, name: str) -> Callable[[Event], Event]:
        """Gets and emitter for a named event.

        Parameters
        ----------
        name :
            The name of the event he requested emitter will emit.
            Users may provide their own named events by requesting an emitter with this function,
            but should do so with caution as it makes time much more difficult to think about.

        Returns
        -------
            An emitter for the named event. The emitter should be called by the requesting component
            at the appropriate point in the simulation lifecycle.
        """
        return self._event_manager.get_emitter(name)

    def register_listener(self, name: str, listener: Callable[[Event], None], priority: int=5) -> None:
        """Registers a callable as a listener to a events with the given name.

        The listening callable will be called with a named ``Event`` as it's only argument any time the
        event emitter is invoked from somewhere in the simulation.

        The framework creates the following events and emits them at different points in the simulation:
            At the end of the setup phase: ``post_setup``
            Every time step: ``time_step__prepare``, ``time_step``, ``time_step__cleanup``, ``collect_metrics``
            At simulation end: ``simulation_end``

        Parameters
        ----------
        name :
            The name of the event to listen for.
        listener :
            The callable to be invoked any time an ``Event`` with the given name is emitted.
        priority : {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
            An indication of the order in which event listeners should be called. Listeners with
            smaller priority values will be called earlier. Listeners with the same priority have
            no guaranteed ordering.  This feature should be avoided if possible. Components should
            strive to obey the Markov property as they transform the state table (the state of the simulation
            at the beginning of the next time step should only depend on the current state of the system).
        """
        self._event_manager.register_listener(name, listener, priority)



import pytest

import pandas as pd
import numpy as np

from vivarium.interpolation import Interpolation, validate_parameters


def test_1d_interpolation():
    df = pd.DataFrame({'a': np.arange(100), 'b': np.arange(100), 'c': np.arange(100, 0, -1)})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), ('a',), 1)

    query = pd.DataFrame({'a': np.arange(100, step=0.01)})

    assert np.allclose(query.a, i(query).b)
    assert np.allclose(100-query.a, i(query).c)


def test_age_year_interpolation():
    years = list(range(1990, 2010))
    ages = list(range(0, 90))
    pops = np.array(ages)*11.1
    data = []
    for age, pop in zip(ages, pops):
        for year in years:
            for sex in ['Male', 'Female']:
                data.append({'age': age, 'sex': sex, 'year': year, 'pop': pop})
    df = pd.DataFrame(data)
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('sex', 'age'), ('year',), 1)

    assert np.allclose(i(year=[1990, 1990], age=[35, 35], sex=['Male', 'Female']), 388.5)


def test_2d_interpolation():
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({'a': a, 'b': b, 'c': b, 'd': a})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), ('a', 'b'), 1)

    query = pd.DataFrame({'a': np.arange(4, step=0.01), 'b': np.arange(4, step=0.01)})

    assert np.allclose(query.b, i(query).c)
    assert np.allclose(query.a, i(query).d)


def test_interpolation_with_categorical_parameters():
    a = ['one']*100 + ['two']*100
    b = np.append(np.arange(100), np.arange(100))
    c = np.append(np.arange(100), np.arange(100, 0, -1))
    df = pd.DataFrame({'a': a, 'b': b, 'c': c})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('a',), ('b',), 1)

    query_one = pd.DataFrame({'a': 'one', 'b': np.arange(100, step=0.01)})
    query_two = pd.DataFrame({'a': 'two', 'b': np.arange(100, step=0.01)})

    assert np.allclose(np.arange(100, step=0.01), i(query_one))

    assert np.allclose(np.arange(100, 0, step=-0.01), i(query_two))


def test_interpolation_with_function():
    df = pd.DataFrame({'a': np.arange(100), 'b': np.arange(100), 'c': np.arange(100, 0, -1)})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), ('a',), 1, func=lambda x: x * 2,)

    query = pd.DataFrame({'a': np.arange(100, step=0.01)})

    assert np.allclose(query.a * 2, i(query).b)


def test_order_zero_2d():
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({'a': a + 0.5, 'b': b + 0.5, 'c': b*3, 'garbage': ['test']*len(a)})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('garbage',), ('a', 'b'), order=0)

    column = np.arange(4, step=0.011)
    query = pd.DataFrame({'a': column, 'b': column, 'garbage': ['test']*(len(column))})

    assert np.allclose(query.b.astype(int) * 3, i(query))


def test_order_zero_1d():
    s = pd.Series({0: 0, 1: 1}).reset_index()
    f = Interpolation(s, tuple(), ('index', ), order=0)

    assert f(index=[0])[0] == 0, 'should be precise at index values'
    assert f(index=[1])[0] == 1
    assert f(index=[2])[0] == 1, 'should be constant extrapolation outside of input range'
    assert f(index=[-1])[0] == 0


def test_validate_parameters__empty_data():
    with pytest.warns(UserWarning) as record:
        out, data = validate_parameters(pd.DataFrame(columns=["age", "sex", "year", "value"]), ["age", "year"], 2)
    assert len(record) == 2
    message = record[0].message.args[0] + " " + record[1].message.args[0]
    assert "age" in message and "year" in message

    assert set(data.columns) == {"sex", "value"}





"""A configuration structure which supports cascading layers.

In Vivarium it allows base configurations to be overridden by component level
configurations which are in turn overridden by model level configuration
which can be overridden by user supplied overrides. From the perspective
of normal client code the cascading is hidden and configuration values
are presented as attributes of the configuration object the values of
which are the value of that key in the outermost layer of configuration
where it appears.

For example:

.. code-block:: python

    >>> config = ConfigTree(layers=['inner_layer', 'middle_layer', 'outer_layer', 'user_overrides'])
    >>> config.read_dict({'section_a': {'item1': 'value1', 'item2': 'value2'}, 'section_b': {'item1': 'value3'}}, layer='inner_layer')
    >>> config.read_dict({'section_a': {'item1': 'value4'}, 'section_b': {'item1': 'value5'}}, layer='middle_layer')
    >>> config.read_dict({'section_b': {'item1': 'value6'}}, layer='outer_layer')
    >>> config.section_a.item1
    'value4'
    >>> config.section_a.item2
    'value2'
    >>> config.section_b.item1
    'value6'
    >>> config.section_b.item1 = 'value7'
    >>> config.section_b.item1
    'value7'
"""
from typing import Mapping, Union
import yaml


class ConfigNode:
    """A single configuration value which may have different variants for different layers.
    Each ConfigNode also records the source (if any is reported) from which the value is derived.
    """
    def __init__(self, layers=None):
        if not layers:
            self._layers = ['base']
        else:
            self._layers = layers
        self._values = {}
        self._frozen = False
        self._accessed = False

    def freeze(self):
        """Causes the node to become read only. This is useful for loading and then
        freezing configurations that should not be modified at runtime.
        """
        self._frozen = True

    def get_value_with_source(self, layer=None):
        """Returns a tuple of the value's source and the value at the specified
        layer. If no layer is specified then the outer layer is used.

        Parameters
        ----------
        layer : str
            Name of the layer to use. If None then the outermost where the value
            exists will be used.

        Raises
        ------
        KeyError
            If the value is not set for the specified layer
        """
        if layer:
            return self._values[layer]

        for layer in reversed(self._layers):
            if layer in self._values:
                return self._values[layer]
        raise KeyError(layer)

    def is_empty(self):
        return not self._values

    def get_value(self, layer=None):
        """Returns the value at the specified layer.

        Parameters
        ----------
        layer : str
            Name of the layer to use. If None then the outermost where the value
            exists will be used.

        Raises
        ------
        KeyError
            If the value is not set for the specified layer
        """
        self._accessed = True
        return self.get_value_with_source(layer)[1]

    def has_been_accessed(self):
        """Returns whether this node has been accessed.

        Returns
        -------
        bool
            Whether this node has been accessed
        """
        return self._accessed

    def metadata(self):
        """Returns all values and associated metadata for this node as a dict.
        The value which would be selected if the node's value was requested
        is indicated by the `default` flag.
        """
        result = []
        for layer in self._layers:
            if layer in self._values:
                result.append({
                    'layer': layer,
                    'value': self._values[layer][1],
                    'source': self._values[layer][0],
                    'default': layer == self._layers[-1]
                })
        return result

    def set_value(self, value, layer=None, source=None):
        """Set a value for a particular layer with optional metadata about source.

        Parameters
        ----------
        value : str
            Data to store in the node.
        layer : str
            Name of the layer to use. If None then the outermost where the value
            exists will be used.
        source : str
            Metadata indicating the source of this value (e.g. a file path)

        Raises
        ------
        TypeError
            If the node is frozen
        KeyError
            If the named layer does not exist
        """
        if self._frozen:
            raise TypeError('Frozen ConfigNode does not support assignment')

        if not layer:
            layer = self._layers[-1]
        self._values[layer] = (source, value)

    def drop_layer(self, layer):
        """Removes the named layer and the value associated with it from the node.

        Parameters
        ----------
        layer : str
            Name of the layer to drop.

        Raises
        ------
        TypeError
            If the node is frozen
        KeyError
            If the named layer does not exist
        """
        if self._frozen:
            raise TypeError('Frozen ConfigNode does not support modification')
        self.reset_layer(layer)
        self._layers.remove(layer)

    def reset_layer(self, layer):
        """Removes any value and metadata associated with the named layer.

        Parameters
        ----------
        layer : str
            Name of the layer to reset.

        Raises
        ------
        TypeError
            If the node is frozen
        KeyError
            If the named layer does not exist
        """
        if self._frozen:
            raise TypeError('Frozen ConfigNode does not support modification')
        if layer in self._values:
            del self._values[layer]

    def __repr__(self):
        return 'ConfigNode(layers={}, values={}, frozen={}, accessed={})'.format(
            self._layers, self._values, self._frozen, self._accessed)

    def __str__(self):
        return '\n'.join(reversed(['{}: {}\n    source: {}'.format(layer, value[1], value[0])
                                   for layer, value in self._values.items()]))


class ConfigTree:
    """A container for configuration information. Each configuration value is
    exposed as an attribute the value of which is determined by the outermost
    layer which has the key defined.
    """
    def __init__(self, data=None, layers=None):
        """
        Parameters
        ----------
        data : dict, str, or ConfigTree, optional
            A dictionary containing initial values
        layers : list
            A list of layer names. The order in which layers defined determines
            how they cascade. Later layers override the values from earlier ones.
        """
        if not layers:
            self.__dict__['_layers'] = ['base']
        else:
            self.__dict__['_layers'] = layers
        self.__dict__['_children'] = {}
        self.__dict__['_frozen'] = False

        if data:
            self.update(data, layer=self._layers[0], source='initial data')

    def freeze(self):
        """Causes the ConfigTree to become read only.

        This is useful for loading and then freezing configurations that should not be modified at runtime.
        """
        self.__dict__['_frozen'] = True
        for child in self._children.values():
            child.freeze()

    def __setattr__(self, name, value):
        """Set a configuration value on the outermost layer."""
        self._set_with_metadata(name, value, layer=None, source=None)

    def __setitem__(self, name, value):
        """Set a configuration value on the outermost layer."""
        self._set_with_metadata(name, value, layer=None, source=None)

    def __getattr__(self, name):
        """Get a configuration value from the outermost layer in which it appears."""
        try:
            return self.get_from_layer(name)
        except KeyError:
            raise AttributeError(name)

    def __getitem__(self, name):
        """Get a configuration value from the outermost layer in which it appears."""
        return self.get_from_layer(name)

    def __delattr__(self, name):
        if name in self._children:
            del self._children[name]

    def __delitem__(self, name):
        if name in self._children:
            del self._children[name]

    def __contains__(self, name):
        """Test if a configuration key exists in any layer."""
        return name in self._children

    def __iter__(self):
        """Dictionary-like iteration."""
        return iter(self._children)

    def items(self):
        """Return a list of all (child_name, child) pairs."""
        return self._children.items()

    def keys(self):
        return self._children.keys()

    def get_from_layer(self, name, layer=None):
        """Get a configuration value from the named layer.

        Parameters
        ----------
        name : str
            The name of the value to retrieve
        layer: str
            The name of the layer to retrieve the value from. If it is not supplied
            then the outermost layer in which the key is defined will be used.
        """
        if name not in self._children:
            if self._frozen:
                raise KeyError(name)
            self._children[name] = ConfigTree(layers=self._layers)
        child = self._children[name]
        if isinstance(child, ConfigNode):
            return child.get_value(layer)
        else:
            return child

    def _set_with_metadata(self, name, value, layer=None, source=None):
        """Set a value in the named layer with the given source.

        Parameters
        ----------
        name : str
            The name of the value
        value
            The value to store
        layer : str, optional
            The name of the layer to store the value in. If none is supplied
            then the value will be stored in the outermost layer.
        source : str, optional
            The source to attribute the value to.

        Raises
        ------
        TypeError
            if the ConfigTree is frozen
        """

        if self._frozen:
            raise TypeError('Frozen ConfigTree does not support assignment')

        if isinstance(value, dict):
            if name not in self._children or not isinstance(self._children[name], ConfigTree):
                self._children[name] = ConfigTree(layers=list(self._layers))
            self._children[name].update(value, layer, source)
        else:
            if name not in self._children or not isinstance(self._children[name], ConfigNode):
                self._children[name] = ConfigNode(list(self._layers))
            child = self._children[name]
            child.set_value(value, layer, source)

    def update(self, data: Union[Mapping, str, bytes], layer: str=None, source: str=None):
        """Adds additional data into the ConfigTree.


        Parameters
        ----------
        data :
            source data
        layer :
            layer to load data into. If none is supplied the outermost one is used
        source :
            Source to attribute the values to

        See Also
        --------
        read_dict
        """
        if isinstance(data, dict):
            self._read_dict(data, layer, source)
        elif isinstance(data, ConfigTree):
            # TODO: set this to parse the other config tree including layer and source info.  Maybe.
            self._read_dict(data.to_dict(), layer, source)
        elif isinstance(data, str):
            if data.endswith('.yaml'):
                source = source if source else data
                self._load(data, layer, source)
            else:
                self._loads(data, layer, source)
        elif data is None:
            pass
        else:
            raise ValueError(f"Update must be called with dictionary, string, or ConfigTree. "
                             f"You passed in {type(data)}")

    def _read_dict(self, data_dict, layer=None, source=None):
        """Load a dictionary into the ConfigTree. If the dict contains nested dicts
        then the values will be added recursively. See module docstring for example code.

        Parameters
        ----------
        data_dict : dict
            source data
        layer : str
            layer to load data into. If none is supplied the outermost one is used
        source : str
            Source to attribute the values to
        """
        for k, v in data_dict.items():
            self._set_with_metadata(k, v, layer, source)

    def _loads(self, data_string, layer=None, source=None):
        """Load data from a yaml formatted string.

        Parameters
        ----------
        data_string : str
            yaml formatted string. The root element of the document should be
            an associative array
        layer : str
            layer to load data into. If none is supplied the outermost one is used
        source : str
            Source to attribute the values to
        """
        data_dict = yaml.load(data_string)
        self._read_dict(data_dict, layer, source)

    def _load(self, f, layer=None, source=None):
        """Load data from a yaml formatted file.

        Parameters
        ----------
        f : str or file like object
            If f is a string then it is interpreted as a path to the file to load
            If it is a file like object then data is read directly from it.
        layer : str
            layer to load data into. If none is supplied the outermost one is used
        source : str
            Source to attribute the values to
        """
        if hasattr(f, 'read'):
            self._loads(f.read(), layer=layer, source=source)
        else:
            with open(f) as f:
                self._loads(f.read(), layer=layer, source=source)

    def to_dict(self):
        result = {}
        for k, v in self._children.items():
            if isinstance(v, ConfigNode):
                result[k] = v.get_value()
            else:
                result[k] = v.to_dict()
        return result

    def metadata(self, name):
        """Return value and metadata associated with the named value

        Parameters
        ----------
        name : str
            name to retrieve. If the name contains '.'s it will be retrieved recursively

        Raises
        ------
        KeyError
            if name is not defined in the ConfigTree
        """
        if name in self._children:
            return self._children[name].metadata()
        else:
            head, _, tail = name.partition('.')
            if head in self._children:
                return self._children[head].metadata(key=tail)
            else:
                raise KeyError(name)

    def reset_layer(self, layer, preserve_keys=()):
        """Removes any value and metadata associated with the named layer.

        Parameters
        ----------
        layer : str
            Name of the layer to reset.
        preserve_keys : list or tuple
            A list of keys to skip while removing data

        Raises
        ------
        TypeError
            If the node is frozen
        KeyError
            If the named layer does not exist
        """
        self._reset_layer(layer, [k.split('.') for k in preserve_keys], prefix=[])

    def _reset_layer(self, layer, preserve_keys, prefix):
        if self._frozen:
            raise TypeError('Frozen ConfigTree does not support modification')
        deletable = []
        for key, child in self._children.items():
            if prefix + [key] not in preserve_keys:
                if isinstance(child, ConfigTree):
                    child._reset_layer(layer, preserve_keys, prefix + [key])
                else:
                    child.reset_layer(layer)
                    if child.is_empty():
                        deletable.append(key)
        for key in deletable:
            del self._children[key]

    def drop_layer(self, layer):
        """Removes the named layer and the value associated with it from the node.

        Parameters
        ----------
        layer : str
            Name of the layer to drop.

        Raises
        ------
        TypeError
            If the node is frozen
        KeyError
            If the named layer does not exist
        """
        if self._frozen:
            raise TypeError('Frozen ConfigTree does not support modification')
        for child in self._children.values():
            child.drop_layer(layer)
        self._layers.remove(layer)

    def unused_keys(self):
        """Lists all keys which are present in the ConfigTree but which have not been accessed."""
        unused = set()
        for k, c in self._children.items():
            if isinstance(c, ConfigNode):
                if not c.has_been_accessed():
                    unused.add(k)
            else:
                for ck in c.unused_keys():
                    unused.add(k+'.'+ck)
        return unused

    def __len__(self):
        return len(self._children)

    def __dir__(self):
        return list(self._children.keys()) + dir(super(ConfigTree, self))

    def __repr__(self):
        return 'ConfigTree(children={}, frozen={})'.format(
            ' '.join([repr(c) for c in self._children.values()]), self._frozen)

    def __str__(self):
        return '\n'.join(['{}:\n    {}'.format(name, str(c).replace('\n', '\n    '))
                          for name, c in self._children.items()])



import pandas as pd

from vivarium.framework.population import PopulationView


class DummyPopulationManager:
    def __init__(self):
        self.population = pd.DataFrame({'age': [0, 10, 20, 30, 40, 50, 60, 70], 'sex': ['Male', 'Female']*4})


def test_create_PopulationView_with_all_columns():
    manager = DummyPopulationManager()
    view = PopulationView(manager)
    assert set(view.columns) == {'age', 'sex'}



import numpy as np
import pandas as pd

from vivarium.testing_utilities import build_table, TestPopulation
from vivarium.interface.interactive import setup_simulation


def test_interpolated_tables(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    ages = build_table(lambda age, sex, year: age, year_start, year_end)
    one_d_age = ages.copy()
    del one_d_age['year']
    one_d_age = one_d_age.drop_duplicates()
    base_config.population.update({'population_size': 10000})

    simulation = setup_simulation([TestPopulation()], input_config=base_config)
    manager = simulation.tables
    years = manager.build_table(years, key_columns=('sex',), parameter_columns=('age', 'year',))
    ages = manager.build_table(ages, key_columns=('sex',), parameter_columns=('age', 'year',))
    one_d_age = manager.build_table(one_d_age, key_columns=('sex',), parameter_columns=('age',))

    result_years = years(simulation.population.population.index)
    result_ages = ages(simulation.population.population.index)
    result_ages_1d = one_d_age(simulation.population.population.index)

    fractional_year = simulation.clock.time.year
    fractional_year += simulation.clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, simulation.population.population.age)
    assert np.allclose(result_ages_1d, simulation.population.population.age)

    simulation.clock._time += pd.Timedelta(30.5 * 125, unit='D')
    simulation.population._population.age += 125/12

    result_years = years(simulation.population.population.index)
    result_ages = ages(simulation.population.population.index)
    result_ages_1d = one_d_age(simulation.population.population.index)

    fractional_year = simulation.clock.time.year
    fractional_year += simulation.clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, simulation.population.population.age)
    assert np.allclose(result_ages_1d, simulation.population.population.age)


def test_interpolated_tables_without_uninterpolated_columns(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    del years['sex']
    years = years.drop_duplicates()
    base_config.population.update({'population_size': 10000})

    simulation = setup_simulation([TestPopulation()], input_config=base_config)
    manager = simulation.tables
    years = manager.build_table(years, key_columns=(), parameter_columns=('year', 'age',))

    result_years = years(simulation.population.population.index)

    fractional_year = simulation.clock.time.year
    fractional_year += simulation.clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)

    simulation.clock._time += pd.Timedelta(30.5 * 125, unit='D')

    result_years = years(simulation.population.population.index)

    fractional_year = simulation.clock.time.year
    fractional_year += simulation.clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)


def test_interpolated_tables__exact_values_at_input_points(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    input_years = years.year.unique()
    base_config.population.update({'population_size': 10000})

    simulation = setup_simulation([TestPopulation()], input_config=base_config)
    manager = simulation.tables
    years = manager.build_table(years, key_columns=('sex',), parameter_columns=('age', 'year',))

    for year in input_years:
        simulation.clock._time = pd.Timestamp(year, 1, 1)
        assert np.allclose(years(simulation.population.population.index),
                           simulation.clock.time.year + 1/365)



#!/usr/bin/env python
import os

from setuptools import setup, find_packages


if __name__ == "__main__":

    base_dir = os.path.dirname(__file__)
    src_dir = os.path.join(base_dir, "src")

    about = {}
    with open(os.path.join(src_dir, "vivarium", "__about__.py")) as f:
        exec(f.read(), about)

    with open(os.path.join(base_dir, "README.rst")) as f:
        long_description = f.read()

    install_requirements = [
        'numpy',
        'pandas',
        'pyyaml',
        'scipy',
        'click',
    ]

    interactive_requirements = [
        'IPython',
        'ipywidgets',
        'jupyter',
        'graphviz',
    ]

    test_requirements = [
        'pytest',
        'pytest-mock',
    ]

    doc_requirements = [
        'sphinx',
        'sphinx-autodoc-typehints',
        'sphinx-rtd-theme',
    ]

    setup(
        name=about['__title__'],
        version=about['__version__'],

        description=about['__summary__'],
        long_description=long_description,
        license=about['__license__'],
        url=about["__uri__"],

        author=about["__author__"],
        author_email=about["__email__"],

        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: BSD",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: Implementation :: CPython",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Life",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Software Development :: Libraries",
        ],

        package_dir={'': 'src'},
        packages=find_packages(where='src'),
        include_package_data=True,

        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            'docs': doc_requirements,
            'test': test_requirements,
            'interactive': interactive_requirements,
            'dev': doc_requirements + test_requirements + interactive_requirements,
        },

        entry_points="""
                [console_scripts]
                simulate=vivarium.interface.cli:simulate
            """,
    )



import warnings

import pandas as pd
from scipy import interpolate


class Interpolation:
    def __init__(self, data, categorical_parameters, continuous_parameters, order, func=None):
        data = data
        self.key_columns = categorical_parameters

        if data.empty:
            raise ValueError("Must supply some input data")

        self.parameter_columns, self_data = validate_parameters(data, continuous_parameters, order)
        self.func = func

        if len(self.parameter_columns) not in [1, 2]:
            raise ValueError("Only interpolation over 1 or 2 variables is supported")

        # These are the columns which the interpolation function will approximate
        value_columns = sorted(data.columns.difference(set(self.key_columns)|set(self.parameter_columns)))
        assert value_columns, (f"No non-parameter data. Avaliable columns: {data.columns}, "
                               f"Parameter columns: {set(self.key_columns)|set(self.parameter_columns)}")

        if self.key_columns:
            # Since there are key_columns we need to group the table by those
            # columns to get the sub-tables to fit
            sub_tables = data.groupby(list(self.key_columns))
        else:
            # There are no key columns so we will fit the whole table
            sub_tables = {None: data}.items()

        self.interpolations = {}

        for key, base_table in sub_tables:
            if base_table.empty:
                continue
            # For each permutation of the key columns build interpolations
            self.interpolations[key] = {}
            for value_column in value_columns:
                # For each value in the table build an interpolation function
                if len(self.parameter_columns) == 2:
                    # 2 variable interpolation
                    if order == 0:
                        x = base_table[list(self.parameter_columns)]
                        y = base_table[value_column]
                        func = interpolate.NearestNDInterpolator(x=x.values, y=y.values)
                    else:
                        index, column = self.parameter_columns
                        table = base_table.pivot(index=index, columns=column, values=value_column)
                        x = table.index.values
                        y = table.columns.values
                        z = table.values
                        func = interpolate.RectBivariateSpline(x=x, y=y, z=z, ky=order, kx=order).ev
                else:
                    # 1 variable interpolation
                    base_table = base_table.sort_values(by=self.parameter_columns[0])
                    x = base_table[self.parameter_columns[0]]
                    y = base_table[value_column]
                    if order == 0:
                        func = interpolate.interp1d(x, y, kind='zero', fill_value='extrapolate')
                    else:
                        func = interpolate.InterpolatedUnivariateSpline(x, y, k=order)
                self.interpolations[key][value_column] = func
        assert self.interpolations

    def __call__(self, *args, **kwargs):
        # TODO: Should be more defensive about this
        if len(args) == 1:
            # We have a dataframe
            df = args[0]
        else:
            # We have parameters for a single invocation
            df = pd.DataFrame(kwargs)

        if self.key_columns:
            sub_tables = df.groupby(list(self.key_columns))
        else:
            sub_tables = [(None, df)]

        result = pd.DataFrame(index=df.index)
        for key, sub_table in sub_tables:
            if sub_table.empty:
                continue
            funcs = self.interpolations[key]
            parameters = tuple(sub_table[k] for k in self.parameter_columns)
            for value_column, func in funcs.items():
                out = func(*parameters)
                # This reshape is necessary because RectBivariateSpline and InterpolatedUnivariateSpline return results
                # in slightly different shapes and we need them to be consistent
                if out.shape:
                    result.loc[sub_table.index, value_column] = out.reshape((out.shape[0],))
                else:
                    result.loc[sub_table.index, value_column] = out

        if self.func:
            return self.func(result)

        if len(result.columns) == 1:
            return result[result.columns[0]]

        return result

    def __repr__(self):
        return "Interpolation()"


def validate_parameters(data, continuous_parameters, order):
    out = []
    for p in continuous_parameters:
        if len(data[p].unique()) > order:
            out.append(p)
        else:
            warnings.warn(f"You requested an order {order} interpolation over the parameter {p}, "
                          f"however there are only {len(data[p].unique())} unique values for {p}"
                          f"which is insufficient to support the requested interpolation order."
                          f"The parameter will be dropped from the interpolation.")
            data = data.drop(p, axis='columns')
    return out, data



"""The engine."""
import gc
import logging
from pprint import pformat
from time import time

import pandas as pd

from vivarium.framework.configuration import build_model_specification
from .components import ComponentInterface
from .event import Event, EventInterface
from .lookup import LookupTableInterface
from .metrics import Metrics
from .plugins import PluginManager
from .population import PopulationInterface
from .randomness import RandomnessInterface
from .results_writer import get_results_writer
from .values import ValuesInterface
from .time import TimeInterface

_log = logging.getLogger(__name__)


class SimulationContext:
    """context"""
    def __init__(self, configuration, components, plugin_manager=None):
        plugin_manager = plugin_manager if plugin_manager else PluginManager()

        self.configuration = configuration
        self.builder = Builder(configuration, plugin_manager)

        self.component_manager = plugin_manager.get_plugin('component_manager')

        self.clock = plugin_manager.get_plugin('clock')
        self.values = plugin_manager.get_plugin('value')
        self.events = plugin_manager.get_plugin('event')
        self.population = plugin_manager.get_plugin('population')
        self.tables = plugin_manager.get_plugin('lookup')
        self.randomness = plugin_manager.get_plugin('randomness')
        for name, controller in plugin_manager.get_optional_controllers().items():
            setattr(self, name, controller)

        # The order the managers are added is important.  It represents the order in which they
        # will be set up.  The clock is required by several of the other managers.  The randomness
        # manager requires the population manager.  The remaining managers need no ordering.
        self.component_manager.add_managers(
            [self.clock, self.population, self.randomness, self.values, self.events, self.tables])
        self.component_manager.add_managers(list(plugin_manager.get_optional_controllers().values()))
        self.component_manager.add_components(components + [Metrics()])

    def setup(self):
        self.component_manager.setup_components(self.builder, self.configuration)
        self.simulant_creator = self.builder.population.get_simulant_creator()

        # The order here matters.
        self.time_step_events = ['time_step__prepare', 'time_step', 'time_step__cleanup', 'collect_metrics']
        self.time_step_emitters = {k: self.builder.event.get_emitter(k) for k in self.time_step_events}
        self.end_emitter = self.builder.event.get_emitter('simulation_end')
        self.builder.event.get_emitter('post_setup')(None)

    def step(self):
        _log.debug(self.clock.time)
        for event in self.time_step_events:
            self.time_step_emitters[event](Event(self.population.population.index))
        self.clock.step_forward()

    def initialize_simulants(self):
        pop_params = self.configuration.population

        # Fencepost the creation of the initial population.
        self.clock.step_backward()
        population_size = pop_params.population_size
        self.simulant_creator(population_size)
        self.clock.step_forward()

    def finalize(self):
        self.end_emitter(Event(self.population.population.index))

    def report(self):
        return self.values.get_value('metrics')(self.population.population.index)

    def __repr__(self):
        return "SimulationContext()"


class Builder:
    """Toolbox for constructing and configuring simulation components."""

    def __init__(self, configuration, plugin_manager):
        self.configuration = configuration

        self.lookup = plugin_manager.get_plugin_interface('lookup')                 # type: LookupTableInterface
        self.value = plugin_manager.get_plugin_interface('value')                   # type: ValuesInterface
        self.event = plugin_manager.get_plugin_interface('event')                   # type: EventInterface
        self.population = plugin_manager.get_plugin_interface('population')         # type: PopulationInterface
        self.randomness = plugin_manager.get_plugin_interface('randomness')         # type: RandomnessInterface
        self.time = plugin_manager.get_plugin_interface('clock')                    # type: TimeInterface
        self.components = plugin_manager.get_plugin_interface('component_manager')  # type: ComponentInterface

        for name, interface in plugin_manager.get_optional_interfaces().items():
            setattr(self, name, interface)

    def __repr__(self):
        return "Builder()"


def run_simulation(model_specification_file, results_directory):
    results_writer = get_results_writer(results_directory, model_specification_file)

    model_specification = build_model_specification(model_specification_file)
    model_specification.configuration.output_data.update(
        {'results_directory': results_writer.results_root}, layer='override', source='command_line')

    simulation = setup_simulation(model_specification)
    metrics, final_state = run(simulation)

    _log.debug(pformat(metrics))
    unused_config_keys = simulation.configuration.unused_keys()
    if unused_config_keys:
        _log.debug("Some configuration keys not used during run: %s", unused_config_keys)

    idx = pd.Index([simulation.configuration.randomness.random_seed], name='random_seed')
    metrics = pd.DataFrame(metrics, index=idx)
    results_writer.write_output(metrics, 'output.hdf')
    results_writer.write_output(final_state, 'final_state.hdf')


def setup_simulation(model_specification):
    plugin_config = model_specification.plugins
    component_config = model_specification.components
    simulation_config = model_specification.configuration

    plugin_manager = PluginManager(plugin_config)
    component_config_parser = plugin_manager.get_plugin('component_configuration_parser')
    components = component_config_parser.get_components(component_config)

    simulation = SimulationContext(simulation_config, components, plugin_manager)
    simulation.setup()
    return simulation


def run(simulation):
    start = time()
    simulation.initialize_simulants()

    while simulation.clock.time < simulation.clock.stop_time:
        gc.collect()  # TODO: Actually figure out where the memory leak is.
        simulation.step()

    simulation.finalize()
    metrics = simulation.report()
    metrics['simulation_run_time'] = time() - start
    return metrics, simulation.population.population



"""System for managing population creation, updating and viewing."""
from typing import Sequence, List, Callable, Union, Mapping, Any
from collections import deque, namedtuple

import pandas as pd

from vivarium import VivariumError


class PopulationError(VivariumError):
    """Error raised when the population system is invalidly queried or updated."""
    pass


class PopulationView:
    """A PopulationView provides access to the simulations population table. It can be used to both read and write
    the state of the population. A PopulationView can only read and write columns for which it is configured. Attempts
    to write to non-existent columns are ignored except during resolution of the ``initialize_simulants`` event when new
    columns are allowed to be created.

    Attributes
    ----------
    columns :
        The columns for which this view is configured. If columns is None then the view will return all columns.
        That case should be only be used in situations where the full state table is actually needed, like some
        metrics collection applications.
    query :
        The query which will be used to filter the population table for this view. This query may reference columns
        not in the view's columns.

    Notes
    -----
    PopulationViews can only be created by the simulation itself. Client code receives them via functions decorated
    by ``uses_columns`` or the builder's ``population_view`` method during setup.
    """

    def __init__(self, manager: 'PopulationManager', columns: Sequence[str]=(), query: str=None):
        self.manager = manager
        self._columns = list(columns)
        self._query = query

    @property
    def columns(self) -> List[str]:
        if not self._columns:
            return list(self.manager.population.columns)
        return list(self._columns)

    def subview(self, columns: Sequence[str]) -> 'PopulationView':
        if set(columns) > set(self._columns):
            raise PopulationError(f"Invalid subview requested.  Requested columns must be a subset of this view's "
                                  f"columns.  Requested columns: {columns}, Available columns: {self.columns}")
        return PopulationView(self.manager, columns, self.query)

    @property
    def query(self) -> str:
        return self._query

    def get(self, index: pd.Index, query: str='', omit_missing_columns: bool=False) -> pd.DataFrame:
        """For the rows in ``index`` get the columns from the simulation's population which this view is configured.
        The result may be further filtered by the view's query.

        Parameters
        ----------
        index :
            Index of the population to get.
        query :
            Conditions used to filter the index.  May use columns not in the requested view.
        omit_missing_columns :
            Silently skip loading columns which are not present in the population table. In general you want this to
            be False because that situation indicates an error but sometimes, like during population initialization,
            it can be convenient to just load whatever data is actually available.

        Returns
        -------
        pd.DataFrame
            A table with the subset of the population requested.
        """

        pop = self.manager.population.loc[index]

        if self._query:
            pop = pop.query(self._query)
        if query:
            pop = pop.query(query)

        if not self._columns:
            return pop
        else:
            if omit_missing_columns:
                columns = list(set(self._columns).intersection(pop.columns))
            else:
                columns = self._columns
            try:
                return pop[columns].copy()
            except KeyError:
                non_existent_columns = set(columns) - set(pop.columns)
                raise PopulationError(f'Requested column(s) {non_existent_columns} not in population table.')

    def update(self, pop: Union[pd.DataFrame, pd.Series]):
        """Update the simulation's state to match ``pop``

        Parameters
        ----------
        pop :
              The data which should be copied into the simulation's state. If ``pop`` is a DataFrame only those columns
              included in the view's columns will be used. If ``pop`` is a Series it must have a name that matches
              one of the view's columns unless the view only has one column in which case the Series will be assumed to
              refer to that regardless of its name.
        """

        if not pop.empty:
            if isinstance(pop, pd.Series):
                if pop.name in self._columns:
                    affected_columns = [pop.name]
                elif len(self._columns) == 1:
                    affected_columns = self._columns
                else:
                    raise PopulationError('Cannot update with a Series unless the series name equals a column '
                                          'name or there is only a single column in the view')
            else:
                affected_columns = set(pop.columns)

            affected_columns = set(affected_columns).intersection(self._columns)
            state_table = self.manager.population
            if not self.manager.growing:
                affected_columns = set(affected_columns).intersection(state_table.columns)

            for c in affected_columns:
                if c in state_table:
                    v = state_table[c].values
                    if isinstance(pop, pd.Series):
                        v2 = pop.values
                    else:
                        v2 = pop[c].values
                    v[pop.index] = v2

                    if v.dtype != v2.dtype:
                        # This happens when the population is being grown because extending
                        # the index forces columns that don't have a natural null type
                        # to become 'object'
                        if not self.manager.growing:
                            raise PopulationError('Component corrupting population table. '
                                                  'Old column type: {} New column type: {}'.format(v.dtype, v2.dtype))
                        v = v.astype(v2.dtype)
                else:
                    if isinstance(pop, pd.Series):
                        v = pop.values
                    else:
                        v = pop[c].values
                self.manager._population[c] = v

    def __repr__(self):
        return "PopulationView(_columns= {}, _query= {})".format(self._columns, self._query)


SimulantData = namedtuple('SimulantData', ['index', 'user_data', 'creation_time', 'creation_window'])


class PopulationManager:
    """The configuration for the population management system.

        Notes
        -----
        Client code should never need to interact with this class
        except through the ``population_view`` function on the builder
        during setup.
    """

    configuration_defaults = {
        'population': {'population_size': 100}
    }

    def __init__(self):
        self._population = pd.DataFrame()
        self._population_initializers = [(self.on_create_simulants, ['tracked'], [])]
        self._initializers_ordered = False
        self.growing = False

    def setup(self, builder):
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()
        builder.value.register_value_modifier('metrics', modifier=self.metrics)

    def get_view(self, columns: Sequence[str], query: str=None) -> PopulationView:
        """Return a configured PopulationView

        Notes
        -----
        Client code should only need this (and only through the version exposed as
        ``population_view`` on the builder during setup) if it uses dynamically
        generated column names that aren't known at definition time. Otherwise
        components should use ``uses_columns``.
        """
        if 'tracked' not in columns:
            query_with_track = query + 'and tracked == True' if query else 'tracked == True'
            return PopulationView(self, columns, query_with_track)
        return PopulationView(self, columns, query)

    def register_simulant_initializer(self, initializer: Callable,
                                      creates_columns: Sequence[str]=(), requires_columns: Sequence[str]=()):
        self._population_initializers.append((initializer, creates_columns, tuple(requires_columns)+('tracked',)))

    def get_simulant_creator(self) -> Callable:
        return self._create_simulants

    def on_create_simulants(self, pop_data):
        status = pd.Series(True, index=pop_data.index)
        self.get_view(['tracked']).update(status)

    def _order_initializers(self) -> None:
        unordered_initializers = deque(self._population_initializers)
        starting_length = -1
        available_columns = []
        self._population_initializers = []

        # This is the brute force N^2 way because constructing a dependency graph is work
        # and in practice this should run in about order N time due to the way dependencies are
        # typically specified.  N is also very small in all current applications.
        while len(unordered_initializers) != starting_length:
            starting_length = len(unordered_initializers)
            for _ in range(len(unordered_initializers)):
                initializer, columns_created, columns_required = unordered_initializers.pop()
                if set(columns_required) <= set(available_columns):
                    self._population_initializers.append((initializer, columns_created, columns_required))
                    available_columns.extend(columns_created)
                else:
                    unordered_initializers.appendleft((initializer, columns_created, columns_required))

        if unordered_initializers:
            raise PopulationError(f"The initializers {unordered_initializers} could not be added.  "
                                  f"Check for cyclic dependencies in your components.")

        if len(set(available_columns)) < len(available_columns):
            raise PopulationError("Multiple components are attempting to initialize the "
                                  "same columns in the state table.")

        self._initializers_ordered = True

    def _create_simulants(self, count: int, population_configuration: Mapping[str, Any]=None) -> pd.Index:
        population_configuration = population_configuration if population_configuration else {}
        if not self._initializers_ordered:
            self._order_initializers()

        new_index = range(len(self._population) + count)
        new_population = self._population.reindex(new_index)
        index = new_population.index.difference(self._population.index)
        self._population = new_population
        self.growing = True
        for initializer, *_ in self._population_initializers:
            initializer(SimulantData(index, population_configuration, self.clock(), self.step_size()))
        self.growing = False
        return index

    def metrics(self, index, metrics):
        population = self.get_view(['tracked']).get(index)
        untracked = population[~population.tracked]
        tracked = population[population.tracked]

        metrics['total_population_untracked'] = len(untracked)
        metrics['total_population_tracked'] = len(tracked)
        metrics['total_population'] = len(untracked)+len(tracked)
        return metrics

    @property
    def population(self) -> pd.DataFrame:
        return self._population.copy()

    def __repr__(self):
        return "PopulationManager()"


class PopulationInterface:

    def __init__(self, population_manager: PopulationManager):
        self._population_manager = population_manager

    def get_view(self, columns: Sequence[str], query: str = None) -> PopulationView:
        """Get a time-varying view of the population state table.

        The requested population view can be used to view the current state or to update the state
        with new values.

        Parameters
        ----------
        columns :
            A subset of the state table columns that will be available in the returned view.
        query :
            A filter on the population state.  This filters out particular rows (simulants) based
            on their current state.  The query should be provided in a way that is understood by
            the ``pandas.DataFrame.query`` method and may reference state table columns not
            requested in the ``columns`` argument.

        Returns
        -------
        PopulationView
            A filtered view of the requested columns of the population state table.
        """
        return self._population_manager.get_view(columns, query)

    def get_simulant_creator(self) -> Callable[[int, Union[Mapping[str, Any], None]], pd.Index]:
        """Grabs a reference to the function that creates new simulants (adds rows to the state table).

        Returns
        -------
        Callable
           The simulant creator function. The creator function takes the number of simulants to be
           created as it's first argument and a dict or other mapping of population configuration
           that will be available to simulant initializers as it's second argument. It generates
           the new rows in the population state table and then calls each initializer
           registered with the population system with a data object containing the state table
           index of the new simulants, the configuration info passed to the creator, the current
           simulation time, and the size of the next time step.
        """
        return self._population_manager.get_simulant_creator()

    def initializes_simulants(self, initializer: Callable[[SimulantData], None],
                              creates_columns: Sequence[str]=(),
                              requires_columns: Sequence[str]=()):
        """Marks a callable as a source of initial state information for new simulants.

        Parameters
        ----------
        initializer :
            A callable that adds or updates initial state information about new simulants.
        creates_columns :
            A list of the state table columns that the given initializer provides the initial state information for.
        requires_columns :
            A list of the state table columns that already need to be present and populated
            in the state table before the provided initializer is called.
        """
        self._population_manager.register_simulant_initializer(initializer, creates_columns, requires_columns)



"""The ``vivarium`` component management system.

This module contains the interface and the default implementation of the ``vivarium`` component manager.
The manager is responsible for tracking all components in the system and for initiating the ``setup``
life-cycle stage of each component.
"""
import inspect
from typing import Sequence

from vivarium import VivariumError


class ComponentConfigError(VivariumError):
    """Error while interpreting configuration file or initializing components"""
    pass


class ComponentManager:
    """Handles the setup and access patterns for components in the system."""

    def __init__(self):
        self._managers = []
        self._components = []

    def add_managers(self, managers: Sequence):
        """Register new managers with the system.

        Managers are setup before components

        Parameters
        ----------
        managers:
          Components to register
        """
        self._add_components(self._managers, managers)

    def add_components(self, components: Sequence):
        """Register new components.

        Parameters
        ----------
        components:
          Components to register
        """
        self._add_components(self._components, components)

    def _add_components(self, component_list, components):
        for component in components:
            if isinstance(component, Sequence):
                self._add_components(component_list, component)
            else:
                component_list.append(component)

    def query_components(self, component_type: str):
        raise NotImplementedError()

    def setup_components(self, builder, configuration):
        """Apply component level configuration defaults to the global config and run setup methods on the components
        registering and setting up any child components generated in the process.

        Parameters
        ----------
        builder:
            Interface to several simulation tools.
        configuration:
            Simulation configuration parameters.
        """
        self._managers = _setup_components(builder, self._managers, configuration)
        self._components = _setup_components(builder, self._components, configuration)


class ComponentInterface:

    def __init__(self, component_manager: ComponentManager):
        self._component_manager = component_manager

    def add_components(self, components: Sequence):
        self._component_manager.add_components(components)

    def query_components(self, component_type: str):
        return self._component_manager.query_components(component_type)


def _setup_components(builder, component_list, configuration):
    done = []
    while component_list:
        component = component_list.pop(0)
        if component is None:
            raise ComponentConfigError('None in component list. This likely '
                                       'indicates a bug in a factory function')
        if component in done:
            continue

        _apply_component_default_configuration(configuration, component)
        if hasattr(component, 'setup'):
            result = component.setup(builder)
            if result is not None:
                # TODO Remove this once we've flushed out all the old style setup methods -Alec 06/05/18
                raise ComponentConfigError("Returning components from setup methods is no longer supported. Use builder.add_components()")
        done.append(component)
    return done


def _apply_component_default_configuration(configuration, component):
    if hasattr(component, 'configuration_defaults'):
        # This reapplies configuration from some components but it is idempotent so there's no effect.
        if component.__module__ == '__main__':
            # This is defined directly in a script or notebook so there's no file to attribute it to
            source = '__main__'
        else:
            source = inspect.getfile(component.__class__)
        configuration.update(component.configuration_defaults, layer='component_configs', source=source)



import os.path

import yaml

from vivarium import VivariumError
from vivarium.config_tree import ConfigTree

from .plugins import DEFAULT_PLUGINS


class ConfigurationError(VivariumError):
    """Error raised when invalid configuration is received."""
    pass


def build_model_specification(model_specification_file_path: str) -> ConfigTree:
    model_specification_file_path = validate_model_specification_file(model_specification_file_path)

    model_specification = _get_default_specification()
    model_specification.update(model_specification_file_path, layer='model_override')

    return model_specification


def validate_model_specification_file(file_path: str) -> str:
    """Ensures the provided file is a yaml file"""
    if not os.path.isfile(file_path):
        raise ConfigurationError('If you provide a model specification file, it must be a file. '
                                 f'You provided {file_path}')

    extension = file_path.split('.')[-1]
    if extension not in ['yaml', 'yml']:
        raise ConfigurationError(f'Model specification files must be in a yaml format. You provided {extension}')
    # Attempt to load
    yaml.load(file_path)
    return file_path


def build_simulation_configuration() -> ConfigTree:
    return _get_default_specification().configuration


def _get_default_specification():
    default_config_layers = ['base', 'component_configs', 'model_override', 'override']
    default_metadata = {'layer': 'base', 'source': 'vivarium_defaults'}

    model_specification = ConfigTree(layers=default_config_layers)
    model_specification.update(DEFAULT_PLUGINS, **default_metadata)
    model_specification.update({'components': None})

    user_config_path = os.path.expanduser('~/vivarium.yaml')
    if os.path.exists(user_config_path):
        model_specification.configuration.update(user_config_path, layer='component_configs')

    return model_specification



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "vivarium"
__summary__ = "vivarium is a microsimulation framework built on top of the standard scientific python stack."
__uri__ = "https://github.com/ihmeuw/vivarium"

__version__ = "0.8.2"

__author__ = "The vivarium developers"
__email__ = "collijk@uw.edu"

__license__ = "GNU GPLv3"
__copyright__ = f"Copyright 2016-2018 {__author__}"
